# Copyright (C) 2018-2023 Mark McIntyre

import csv
from matplotlib import pyplot as plt
import numpy as np

from ukmon_meteortools.utils import greatCircleDistance


def trackToDistvsHeight(trackcsvfile):
    """
    Plot a distance vs height graph from the supplied CSV file  

    Arguments:  
        trackcsvfile:   [str] full path to a CSV file containing columns of lat, long, height, time  

    Returns:  
        nothing, but it creates a PNG in the source folder containing the track plot  
    """
    inputfile = csv.reader(open(trackcsvfile))
    dists = []
    alts = []
    lat0 = -99   # use impossible value
    lng0 = 0
    for row in inputfile:
        #columns are lat, long, height, times
        if row[0] == 'lats': 
            continue
        if lat0 == -99:
            lat0 = np.radians(float(row[0]))
            lng0 = np.radians(float(row[1]))
            dist = 0
        else:
            lat = np.radians(float(row[0]))
            lng = np.radians(float(row[1]))
            dist = greatCircleDistance(lat0, lng0, lat, lng)
        dists.append(dist)
        alt = float(row[2])/1000
        alts.append(alt)
    plt.clf()
    plt.plot(dists,alts)
    outname = trackcsvfile.replace('.csv','_dist_alt.png')
    plt.savefig(outname)
    plt.close()


def trackToTimevsVelocity(trackcsvfile):
    """
    Plot a distance vs velocity graph from the supplied CSV file  

    Arguments:  
        trackcsvfile:   [str] full path to a CSV file containing columns of lat, long, height, time  

    Returns:  
        nothing, but it creates a PNG in the source folder containing the track plot  
    """
    inputfile = csv.reader(open(trackcsvfile))
    dists = []
    tims = []
    lat0 = -99   # use impossible value
    lng0 = 0
    for row in inputfile:
        #columns are lat, long, height, times
        if row[0] == 'lats': 
            continue
        if lat0 == -99:
            lat0 = np.radians(float(row[0]))
            lng0 = np.radians(float(row[1]))
            dist = 0
        else:
            lat = np.radians(float(row[0]))
            lng = np.radians(float(row[1]))
            dist = greatCircleDistance(lat0, lng0, lat, lng)
        dists.append(dist)
        tim = float(row[3])
        tims.append(tim)
    vels=[]
    vels.append(0)
    for i in range(1,len(tims)):
        vel = (dists[i]-dists[i-1])/(tims[i]-tims[i-1])
        print(vel)
        vels.append(vel)
    plt.clf()
    plt.plot(tims,vels)
    outname = trackcsvfile.replace('.csv','_tim_vel.png')
    plt.savefig(outname)
    plt.close()


def trackToTimevsHeight(trackcsvfile):
    """
    Plot a time vs height graph from the supplied CSV file  

    Arguments:  
        trackcsvfile:   [str] full path to a CSV file containing columns of lat, long, height, time  

    Returns:  
        nothing, but it creates a PNG in the source folder containing the track plot  
    """
    inputfile = csv.reader(open(trackcsvfile))
    tims = []
    alts = []
    for row in inputfile:
        #columns are lat, long, height, times
        if row[0] == 'lats': 
            continue
        tims.append(float(row[3]))
        alts.append(float(row[2])/1000)
    plt.clf()
    plt.plot(tims,alts)
    outname = trackcsvfile.replace('.csv','_time_alt.png')
    plt.savefig(outname)
    plt.close()



#
# Load an FTPdetectInfo file - copied from RMS
#
# Copyright (C) 2018-2023 Mark McIntyre

import os
import numpy as np
import configparser as crp
import json
import datetime

from ukmon_meteortools.utils import date2JD, angleBetweenSphericalCoords


def filterFTPforSpecificTime(ftpfile, dtstr):
    """ filter FTPdetect file for a specific event, by time, and copy it into a new file  

    Arguments:  
        ftpfile - [string] full path to the ftpdetect file to filter  
        dtstr   - [string] date/time of required event in yyyymmdd_hhmmss format  

    Returns:  
        tuple containing  
            full name of the new file containing just matching events  
            the number of matching events  
    """
    meteor_list = loadFTPDetectInfo(ftpfile, time_offsets=None, join_broken_meteors=True, locdata=None)
    refdt = datetime.datetime.strptime(dtstr, '%Y%m%d_%H%M%S')
    #print(refdt)
    new_met_list = []
    for met in meteor_list:
        #print(met.ff_name)
        dtpart = datetime.datetime.strptime(met.ff_name[10:25], '%Y%m%d_%H%M%S')
        tdiff = (refdt - dtpart).seconds
        #print(tdiff)
        if abs(tdiff) < 21:
            print('adding one entry')
            new_met_list.append(met)
    newname = writeNewFTPFile(ftpfile, new_met_list)
    return newname, len(new_met_list)


def writeNewFTPFile(srcname, metlist):
    """ creates a FTPDetect file from a list of MeteorObservation objects  
        
    Arguments: 
        srcname     - [string] full path to the original FTP file  
        metlist     - list of MeteorObservation objects  

    Returns: 
        the full path to the created file 

    """
    outdir, fname = os.path.split(srcname)
    newname = os.path.join(outdir, f'{fname}.old')
    try:
        os.rename(srcname, newname)
    except:
        pass
    if os.path.isfile(srcname):
        srcname = srcname[:-4] + '_new.txt'
    with open(srcname, 'w') as ftpf:
        _writeFTPHeader(ftpf, len(metlist), outdir, False)
        metno = 1
        ffname = ''
        for met in metlist:
            if ffname == met.ff_name:
                metno = metno + 1
            else:
                metno = 1
                ffname = met.ff_name
            _writeOneMeteor(ftpf, metno, met.station_id, met.time_data, len(met.frames), met.fps, met.frames, 
                np.degrees(met.ra_data), np.degrees(met.dec_data), 
                np.degrees(met.azim_data), np.degrees(met.elev_data),
                None, met.mag_data, False, met.x_data, met.y_data, met.ff_name)
    return srcname


def loadFTPDetectInfo(ftpdetectinfo_file_name, time_offsets=None,
        join_broken_meteors=True, locdata=None):
    """ Loads an FTPDEtect file into a list of MeteorObservation objects  

    Arguments:  
        ftpdetectinfo_file_name: [str] Path to the FTPdetectinfo file.  
        stations: [dict] A dictionary where the keys are stations IDs, and values are lists of:  
            - latitude +N in radians  
            - longitude +E in radians  
            - height in meters  
        

    Keyword arguments:  
        time_offsets: [dict] (key, value) pairs of (stations_id, time_offset) for every station. None by 
            default.
        join_broken_meteors: [bool] Join meteors broken across 2 FF files.


    Return:  
        meteor_list: [list] A list of MeteorObservation objects filled with data from the FTPdetectinfo file.  

    """
    stations={}
    if locdata is None:
        dirname, fname = os.path.split(ftpdetectinfo_file_name)
        cfgfile = os.path.join(dirname, '.config')
        cfg = crp.ConfigParser()
        cfg.read(cfgfile)
        try: 
            lat = float(cfg['System']['latitude'].split()[0])
            lon = float(cfg['System']['longitude'].split()[0])
            height = float(cfg['System']['elevation'].split()[0])
        except:
            # try reading from platepars file
            ppf = os.path.join(dirname, 'platepars_all_recalibrated.json')
            if not os.path.isfile(ppf):
                return []
            js = json.load(open(ppf, 'r'))
            if len(js) < 10:
                return []
            lat = js[list(js.keys())[0]]['lat']
            lon = js[list(js.keys())[0]]['lon']
            height = js[list(js.keys())[0]]['elev']
        statid= fname.split('_')[1]
    else:
        statid = locdata['station_code']
        lat = float(locdata['lat'])
        lon = float(locdata['lon'])
        height = float(locdata['elev'])

    stations[statid] = [np.radians(lat), np.radians(lon), height*1000]
    meteor_list = []

    with open(ftpdetectinfo_file_name) as f:
        # Skip the header
        for i in range(11):
            next(f)

        current_meteor = None

        bin_name = False
        cal_name = False
        meteor_header = False

        for line in f:
            # Skip comments
            if line.startswith("#"):
                continue

            line = line.replace('\n', '').replace('\r', '')

            # Skip the line if it is empty
            if not line:
                continue

            if '-----' in line:
                # Mark that the next line is the bin name
                bin_name = True

                # If the separator is read in, save the current meteor
                if current_meteor is not None:
                    current_meteor._finish()
                    meteor_list.append(current_meteor)
                continue

            if bin_name:
                bin_name = False

                # Mark that the next line is the calibration file name
                cal_name = True

                # Save the name of the FF file
                ff_name = line

                # Extract the reference time from the FF bin file name
                line = line.split('_')

                # Count the number of string segments, and determine if it the old or new CAMS format
                if len(line) == 6:
                    sc = 1
                else:
                    sc = 0

                ff_date = line[1 + sc]
                ff_time = line[2 + sc]
                milliseconds = line[3 + sc]

                year = ff_date[:4]
                month = ff_date[4:6]
                day = ff_date[6:8]

                hour = ff_time[:2]
                minute = ff_time[2:4]
                seconds = ff_time[4:6]

                year, month, day, hour, minute, seconds, milliseconds = map(int, [year, month, day, hour, 
                    minute, seconds, milliseconds])

                # Calculate the reference JD time
                jdt_ref = date2JD(year, month, day, hour, minute, seconds, milliseconds)
                continue

            if cal_name:
                cal_name = False
                # Mark that the next line is the meteor header
                meteor_header = True
                continue

            if meteor_header:
                meteor_header = False
                line = line.split()

                # Get the station ID and the FPS from the meteor header
                station_id = line[0].strip()
                fps = float(line[3])

                # Try converting station ID to integer
                try:
                    station_id = int(station_id)
                except:
                    pass

                # If the time offsets were given, apply the correction to the JD
                if time_offsets is not None:
                    if station_id in time_offsets:
                        print('Applying time offset for station {:s} of {:.2f} s'.format(str(station_id),
                            time_offsets[station_id]))

                        jdt_ref += time_offsets[station_id]/86400.0
                    else:
                        print('Time offset for given station not found!')

                # Get the station data
                if station_id in stations:
                    lat, lon, height = stations[station_id]
                else:
                    print('ERROR! No info for station ', station_id, ' found in CameraSites.txt file!')
                    print('Exiting...')
                    break
                # Init a new meteor observation
                current_meteor = MeteorObservation(jdt_ref, station_id, lat, lon, height, fps,
                    ff_name=ff_name)
                continue

            # Read in the meteor observation point
            if (current_meteor is not None) and (not bin_name) and (not cal_name) and (not meteor_header):

                line = line.replace('\n', '').split()

                # Read in the meteor frame, RA and Dec
                frame_n = float(line[0])
                x = float(line[1])
                y = float(line[2])
                ra = float(line[3])
                dec = float(line[4])
                azim = float(line[5])
                elev = float(line[6])

                # Read the visual magnitude, if present
                if len(line) > 8:
                    mag = line[8]
                    if mag == 'inf':
                        mag = None
                    else:
                        mag = float(mag)
                else:
                    mag = None

                # Add the measurement point to the current meteor 
                current_meteor._addPoint(frame_n, x, y, azim, elev, ra, dec, mag)

        # Add the last meteor the the meteor list
        if current_meteor is not None:
            current_meteor._finish()
            meteor_list.append(current_meteor)

    # Concatenate observations across different FF files ###
    if join_broken_meteors:

        # Go through all meteors and compare the next observation
        merged_indices = []
        for i in range(len(meteor_list)):

            # If the next observation was merged, skip it
            if (i + 1) in merged_indices:
                continue

            # Get the current meteor observation
            met1 = meteor_list[i]

            if i >= (len(meteor_list) - 1):
                break

            # Get the next meteor observation
            met2 = meteor_list[i + 1]
            
            # Compare only same station observations
            if met1.station_id != met2.station_id:
                continue

            # Extract frame number
            met1_frame_no = int(met1.ff_name.split("_")[-1].split('.')[0])
            met2_frame_no = int(met2.ff_name.split("_")[-1].split('.')[0])

            # Skip if the next FF is not exactly 256 frames later
            if met2_frame_no != (met1_frame_no + 256):
                continue


            # Check for frame continouty
            if (met1.frames[-1] < 254) or (met2.frames[0] > 2):
                continue

            # Check if the next frame is close to the predicted position 

            # Compute angular distance between the last 2 points on the first FF
            ang_dist = angleBetweenSphericalCoords(met1.dec_data[-2], met1.ra_data[-2], met1.dec_data[-1],
                met1.ra_data[-1])

            # Compute frame difference between the last frame on the 1st FF and the first frame on the 2nd FF
            df = met2.frames[0] + (256 - met1.frames[-1])

            # Skip the pair if the angular distance between the last and first frames is 2x larger than the 
            #   frame difference times the expected separation
            ang_dist_between = angleBetweenSphericalCoords(met1.dec_data[-1], met1.ra_data[-1],
                met2.dec_data[0], met2.ra_data[0])

            if ang_dist_between > 2*df*ang_dist:
                continue

            # If all checks have passed, merge observations ###
            # Recompute the frames
            frames = 256.0 + met2.frames

            # Recompute the time data
            time_data = frames/met1.fps

            # Add the observations to first meteor object
            met1.frames = np.append(met1.frames, frames)
            met1.time_data = np.append(met1.time_data, time_data)
            met1.x_data = np.append(met1.x_data, met2.x_data)
            met1.y_data = np.append(met1.y_data, met2.y_data)
            met1.azim_data = np.append(met1.azim_data, met2.azim_data)
            met1.elev_data = np.append(met1.elev_data, met2.elev_data)
            met1.ra_data = np.append(met1.ra_data, met2.ra_data)
            met1.dec_data = np.append(met1.dec_data, met2.dec_data)
            met1.mag_data = np.append(met1.mag_data, met2.mag_data)

            # Merge the FF file name and create a list
            if (met1.ff_name is not None) and (met2.ff_name is not None):
                met1.ff_name = met1.ff_name + ',' + met2.ff_name

            # Sort all observations by time
            met1._finish()

            # Indicate that the next observation is to be skipped
            merged_indices.append(i + 1)

        # Removed merged meteors from the list
        meteor_list = [element for i, element in enumerate(meteor_list) if i not in merged_indices]

    return meteor_list


class MeteorObservation(object):
    """ Container for meteor observations.  
    """
    def __init__(self, jdt_ref, station_id, latitude, longitude, height, fps, ff_name=None):
        """ Construct the MeteorObservation object.  
        Arguments:  
            jdt_ref: [float] Reference Julian date when the relative time is t = 0s.  
            station_id: [str] Station ID.  
            latitude: [float] Latitude +N in radians.  
            longitude: [float] Longitude +E in radians.  
            height: [float] Elevation above sea level (MSL) in meters. 
            fps: [float] Frames per second. 

        Keyword arguments:  
            ff_name: [str] Name of the originating FF file. 
        """
        self.jdt_ref = jdt_ref
        self.station_id = station_id
        self.latitude = latitude
        self.longitude = longitude
        self.height = height
        self.fps = fps
        self.ff_name = ff_name
        self.frames = []
        self.time_data = []
        self.x_data = []
        self.y_data = []
        self.azim_data = []
        self.elev_data = []
        self.ra_data = []
        self.dec_data = []
        self.mag_data = []
        self.abs_mag_data = []

    def _addPoint(self, frame_n, x, y, azim, elev, ra, dec, mag):
        """ Adds the measurement point to the meteor.

        Arguments:
            frame_n: [flaot] Frame number from the reference time.
            x: [float] X image coordinate.
            y: [float] X image coordinate.
            azim: [float] Azimuth, J2000 in degrees.
            elev: [float] Elevation angle, J2000 in degrees.
            ra: [float] Right ascension, J2000 in degrees.
            dec: [float] Declination, J2000 in degrees.
            mag: [float] Visual magnitude.
        """
        self.frames.append(frame_n)

        # Calculate the time in seconds w.r.t. to the reference JD
        point_time = float(frame_n)/self.fps
        self.time_data.append(point_time)
        self.x_data.append(x)
        self.y_data.append(y)

        # Angular coordinates converted to radians
        self.azim_data.append(np.radians(azim))
        self.elev_data.append(np.radians(elev))
        self.ra_data.append(np.radians(ra))
        self.dec_data.append(np.radians(dec))
        self.mag_data.append(mag)

    def _finish(self):
        """ When the initialization is done, convert data lists to numpy arrays. """
        self.frames = np.array(self.frames)
        self.time_data = np.array(self.time_data)
        self.x_data = np.array(self.x_data)
        self.y_data = np.array(self.y_data)
        self.azim_data = np.array(self.azim_data)
        self.elev_data = np.array(self.elev_data)
        self.ra_data = np.array(self.ra_data)
        self.dec_data = np.array(self.dec_data)
        self.mag_data = np.array(self.mag_data)
        # Sort by frame
        temp_arr = np.c_[self.frames, self.time_data, self.x_data, self.y_data, self.azim_data, 
        self.elev_data, self.ra_data, self.dec_data, self.mag_data]
        temp_arr = temp_arr[np.argsort(temp_arr[:, 0])]
        self.frames, self.time_data, self.x_data, self.y_data, self.azim_data, self.elev_data, self.ra_data, \
            self.dec_data, self.mag_data = temp_arr.T


def _writeFTPHeader(ftpf, metcount, fldr, ufo=True):
    """
    Internal function to create the header of the FTPDetect file  
    """
    l1 = 'Meteor Count = {:06d}\n'.format(metcount)
    ftpf.write(l1)
    ftpf.write('-----------------------------------------------------\n')
    if ufo is True:
        ftpf.write('Processed with UFOAnalyser\n')
    else:
        ftpf.write('Processed with RMS 1.0\n')
    ftpf.write('-----------------------------------------------------\n')
    l1 = 'FF  folder = {:s}\n'.format(fldr)
    ftpf.write(l1)
    l1 = 'CAL folder = {:s}\n'.format(fldr)
    ftpf.write(l1)
    ftpf.write('-----------------------------------------------------\n')
    ftpf.write('FF  file processed\n')
    ftpf.write('CAL file processed\n')
    ftpf.write('Cam# Meteor# #Segments fps hnr mle bin Pix/fm Rho Phi\n')
    ftpf.write('Per segment:  Frame# Col Row RA Dec Azim Elev Inten Mag\n')


def _writeOneMeteor(ftpf, metno, sta, evttime, fcount, fps, fno, ra, dec, az, alt, b, mag, 
        ufo=True, x=None, y=None, ffname = None):
    """   Internal function to write one meteor event into the file in FTPDetectInfo style
    """
    ftpf.write('-------------------------------------------------------\n')
    if ffname is None:
        ms = '{:03d}'.format(int(evttime.microsecond / 1000))

        fname = 'FF_' + sta + '_' + evttime.strftime('%Y%m%d_%H%M%S_') + ms + '_0000000.fits\n'
    else:
        fname = ffname + '\n'
    ftpf.write(fname)

    if ufo is True:
        ftpf.write('UFO UKMON DATA Recalibrated on: ')
    else:
        ftpf.write('RMS data reprocessed on: ')
    ftpf.write(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f UTC\n'))
    li = f'{sta} {metno:04d} {fcount:04d} {fps:04.2f} 000.0 000.0  00.0 000.0 0000.0 0000.0\n'
    ftpf.write(li)

    for i in range(len(fno)):
        #    204.4909 0422.57 0353.46 262.3574 +16.6355 267.7148 +23.0996 000120 3.41
        bri = 0
        if b is not None:
            bri = int(b[i])
        if ufo is True:
            # UFO is timestamped as at the first detection
            thisfn = fno[i] - fno[0]
            thisx = 0
            thisy = 0
        else:
            thisfn = fno[i]
            thisx = x[i]
            thisy = y[i]
        li = f'{thisfn:.4f} {thisx:07.2f} {thisy:07.2f} '
        li = li + f'{ra[i]:8.4f} {dec[i]:+7.4f} {az[i]:8.4f} '
        li = li + f'{alt[i]:+7.4f} {bri:06d} {mag[i]:.02f}\n'
        ftpf.write(li)



# Copyright (C) 2018-2023 Mark McIntyre

# vector maths for comparing directions of meteors etc
import numpy as np
import math


def shortestDistance(a, b, c, d):
    # Lines are L=a+bt, M=c+ds
    e = a - c
    b2 = np.dot(b, b)
    d2 = np.dot(b, b)
    bd = np.dot(b, d)
    de = np.dot(d, e)
    be = np.dot(b, e)
    db = np.dot(d, b)

    A = -b2 * d2 + bd * bd
    s = (-b2 * de + be * db) / A
    t = (+d2 * be - de * db) / A
    D = e + b * t - d * s
    d = math.sqrt(np.dot(D, D))

    return s, t, d



# 
# class to read a single or multiple platepar files
#
# Copyright (C) 2018-2023 Mark McIntyre

import json
import os
import glob


class platepar:
    """ Load and manage an RMS platepar object, used to map an image to the sky """
    def __init__(self, plateparfile):
        """" Create and load a platepar object from a file plateparfile """
        with open(plateparfile, 'r') as pp:
            self.jsd = json.load(pp)
            try: 
                self.id = self.jsd['station_code']
                # field of view details
                self.alt_centre = self.jsd['alt_centre']
                self.az_centre = self.jsd['az_centre']
                self.Ho = self.jsd['Ho']
                self.fov_h = self.jsd['fov_h']
                self.fov_v = self.jsd['fov_v']
                self.rotation_from_horiz = self.jsd['rotation_from_horiz']
                self.RA_d = self.jsd['RA_d']
                self.dec_d = self.jsd['dec_d']
                self.pos_angle_ref = self.jsd['pos_angle_ref']

                # camera location
                self.lat = self.jsd['lat']
                self.lon = self.jsd['lon']
                self.elev = self.jsd['elev']

                # camera resolution
                self.X_res = self.jsd['X_res']
                self.Y_res = self.jsd['Y_res']

                # Reference time and date
                self.time = 0
                self.JD = self.jsd['JD']
                self.UT_corr =self.jsd['UT_corr']

                # FOV scale (px/deg)
                self.F_scale = self.jsd['F_scale']

            except:
                print('invalid json file')

    def getLocation(self):
        """ get the lat, lon and elevation from the platepar"""
        return [self.lat, self.lon, self.elev]

    def getView(self):
        """ get the pointing direction (alt/az), field of view (h/v) and rotation """
        return [self.altcentre, self.azcentre, self.fov_h, self.fov_v, self.rot]

    def getResolution(self):
        """ Get the reported camera resolution """
        return [self.xres, self.yres]


def loadPlatepars(fldr):
    """ Create a dictionary of multiple platepars  
    
    Arguments:  
        fldr:   [string] A folder containing one or more platepar files  

    Returns:  
        a json dict containing each platepar indexed by the corresponding filename  
    """
    platepars ='{'
    pth = os.path.join(fldr, '*.json')
    pps = glob.glob(pth)
    for pp in pps:
        thispl = platepar(pp)
        if thispl is not None:
            if len(platepars) > 1: # and pp != pps[-1]:
                platepars = platepars + ','
            platepars = platepars + '"' + thispl.id +'": '+ json.dumps(thispl.jsd)
    platepars = platepars + '}'
    #print(platepars)
    ppjson = json.loads(platepars)
    return ppjson



import pickle
import tempfile
import pandas as pd
import numpy as np
import requests
import os
import json
# imports are required to load the picklefiles
try:
    from wmpl.Trajectory.CorrelateRMS import TrajectoryReduced, DatabaseJSON # noqa: F401
    from wmpl.Trajectory.CorrelateRMS import MeteorObsRMS, PlateparDummy, MeteorPointRMS # noqa: F401
except:
    print('WMPL not available')

from ukmon_meteortools.usertools.getLiveImages import _download
from ukmon_meteortools.fileformats import trackCsvtoKML


def getTrajPickle(trajname):
    """ Retrieve a the pickled trajectory for a matched detection 
    
    Arguments:  
        trajname:   [string] Name of the trajectory eg 20230502_025228.374_UK_BE

    Returns:
        WMPL traj object

    Notes:
        WMPL must be in the $PYTHONPATH for this function to be available

    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    fmt = 'pickle'
    apicall = f'{apiurl}?reqval={trajname}&format={fmt}'
    matchpickle = pd.read_json(apicall, lines=True)
    if 'url' in matchpickle:
        outdir = tempfile.mkdtemp()
        _download(matchpickle.url[0], outdir, matchpickle.filename[0])

        localfile = os.path.join(outdir, matchpickle.filename[0])
        with open(localfile, 'rb') as f:
            traj = pickle.load(f, encoding='latin1')
        try:
            os.remove(localfile)
        except:
            print(f'unable to remove {localfile}')
    else:
        traj = None
    return traj


def trajectoryKML(trajname, outdir=None):
    """
    Create a 3-d KML file showing the trajectory of a meteor, which can be 
    loaded into Google Maps. 
    
    Arguments:  
        trajname:   [string] trajectory name eg 20230213_025913.678_UK
        outdir:     [string] where to save the file. Defaults to current location  

    Returns:  
        kmlfile:    name of the saved KML file  
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    fmt = 'json'
    apicall = f'{apiurl}?reqval={trajname}&format={fmt}'
    res = requests.get(apicall, stream=True)
    if res.status_code == 200:
        data=b''
        for chunk in res.iter_content(chunk_size=4096):
            data = data + chunk
        jsd = json.loads(data)    
        lats = []
        lons = []
        hts = [] 
        rngs = []
        for js in jsd['observations']:
            k = list(js.keys())[0]
            thisstat = js[k]
            lats = lats + thisstat['model_lat']
            lons = lons + thisstat['model_lon']
            hts = hts + thisstat['model_ht']
            rngs = rngs + thisstat['model_range']
        df = pd.DataFrame({"lats": np.degrees(lats), "lons": np.degrees(lons), "alts": hts, "ranges": rngs})
        df = df.sort_values(by=['ranges', 'lats'])
        outfname = os.path.join(outdir, f'{trajname}.kml')
        if outdir is None:
            outdir = '.'
            os.makedirs(outdir, exist_ok=True)
        trackCsvtoKML(trajname, df, saveOutput=True, outdir=outdir)
        return outfname
    else:
        print('no match')
        return None



# Copyright (C) 2018-2023 Mark McIntyre
#
# simple script to get the active shower list from the IMO working list

import numpy as np
import sys
import os
import datetime

from ukmon_meteortools.utils import jd2Date, sollon2jd
try:
    from UpdateOrbitFiles import updateOrbitFiles ## to update WMPL raw data files
    gotupdater = True
except:
    gotupdater = False


def _loadFullData(pth=None):
    return _loadDataFile(1, pth)


def _loadLookupTable(pth=None):
    return _loadDataFile(2, pth)


def _loadJenniskensShowers(dir_path, file_name):
    """ Load the showers from the Jenniskens et al. (2018) table and init MeteorShower objects. """
    jenniskens_shower_list = []
    with open(os.path.join(dir_path, file_name), encoding='cp1252') as f:
        data_start = 0
        for line in f:
            if "====================================" in line:
                data_start += 1
                continue
            if data_start < 2:
                continue
            line = line.replace('\n', '').replace('\r', '')
            if not line:
                continue
            if "[FINAL]" in line:
                break
            l0, L_l0, B_g, v_g, IAU_no = line.split()
            jenniskens_shower_list.append([np.radians(float(l0)), np.radians(float(L_l0)), 
                np.radians(float(B_g)), 1000*float(v_g), int(IAU_no)])
    return np.array(jenniskens_shower_list)


def _refreshShowerData():
    if gotupdater is True:
        updateOrbitFiles()
    abs_path = os.getenv('WMPL_LOC', default='/home/ec2-user/src/WesternMeteorPyLib')
    jenniskens_shower_table_file = os.path.join(abs_path, 'wmpl', 'share', 'ShowerLookUpTable.txt')
    jenniskens_shower_table_npy = os.path.join(abs_path, 'wmpl', 'share', 'ShowerLookUpTable.npy')
    jenniskens_shower_list = _loadJenniskensShowers(*os.path.split(jenniskens_shower_table_file))
    np.save(jenniskens_shower_table_npy, jenniskens_shower_list)
    iau_shower_table_file = os.path.join(abs_path, 'wmpl', 'share', 'streamfulldata.csv')
    iau_shower_table_npy = os.path.join(abs_path, 'wmpl', 'share', 'streamfulldata.npy')
    iau_shower_list = np.loadtxt(iau_shower_table_file, delimiter="|", usecols=range(20), dtype=str)
    np.save(iau_shower_table_npy, iau_shower_list)


def _loadDataFile(typ, pth=None):
    if typ == 1:
        fname='streamfulldata.npy'
    elif typ == 2:
        fname='ShowerLookupTable.npy'
    else:
        return 'invalid type code'

    if pth is None:
        if sys.platform == 'win32':
            pth = 'e:/dev/meteorhunting/WesternMeteorPyLib/wmpl/share'
        else:
            pth = '/home/ec2-user/src/WesternMeteorPyLib/wmpl/share'
    dfil = np.load(os.path.join(pth, fname))
    return dfil


def getShowerDets(shwr, stringFmt=False):
    """ Get details of a shower 
    
    Arguments:  
        shwr:   [string] three-letter shower code eg PER  
         
    Returns:  
        (id, full name, peak solar longitude, peak date mm-dd)  
    """
    sfd = _loadFullData()
    sfdfltr = sfd[sfd[:,3] == shwr]
    mtch = [sh for sh in sfdfltr if sh[6] != '-2']
    if len(mtch) > 0:
        id = int(mtch[-1][1])
        nam = mtch[-1][4].strip()
        pksollong = float(mtch[-1][7])
        dt = datetime.datetime.now()
        yr = dt.year
        mth = dt.month
        jd = sollon2jd(yr, mth, pksollong)
        pkdt = jd2Date(jd, dt_obj=True)
        dtstr = pkdt.strftime('%m-%d')
    else:
        id, nam, pksollong, dtstr = 0, 'Unknown', 0, 'Unknown'
    if stringFmt:
        return f"{pksollong},{dtstr},{nam},{shwr}"
    else:
        return id, nam, pksollong, dtstr


def getShowerPeak(shwr):
    """ Get date of a shower peak in MM-DD format
    
    Arguments:  
        shwr:   [string] three-letter shower code eg PER  
         
    Returns:  
        peak date mm-dd  
    """
    _, _, _, pk = getShowerDets(shwr)
    return pk



# Copyright (C) 2018-2023 Mark McIntyre

# python script to call TrackStack to stack several nights/cameras

import os
import glob
import datetime
import argparse
try:
    from Utils.TrackStack import trackStack
    import RMS.ConfigReader as cr
except Exception:
    print('RMS not available')


def multiTrackStack(camlist, start, end, outdir=None, shwr=None, scale=None, draw_cons=False, noplot=True, datadir=None):
    """
    create a multi camera and/or multi night trackstack  

    Arguments:  
        camlist:    [list] list of camera IDs eg ['UK0006','UK000F']  
        start:      [string] start date as a string yyyymmdd  
        end:        [string] end date as a string yyyymmdd  
    
    Keyword Arguments: 
        outdir:     [string] where to save to. Default current working directory.  
        shwr:       [string] filter for a specific shower eg 'PER'. Default all showers.  
        scale:      [int] scale the image to avoid cropping. Default 1.  
        draw_cons:  [bool] Draw constellation stick figures. Default false.  
        noplot:     [bool] Don't display the plot, just save it. Default true.  
        datadir:    [string] Root of where to read the data from.   

    Notes:  
        The function expects data to be stored in RMS folder structures as follows  
            {datadir}/{cameraid}/ConfirmedFiles/{cameraid_date_time_*}  

        Each folder must contain an FTPdetectinfo file and a platepars_all file. 
        The first-named folder must contain a valid RMS config file.  

        It is assumed that all cameras are at the same location. The output for cameras 
        at different locations has not been tested.  

        If you find the image is being cropped, try increasing scale. This will use more memory.  

    """
    if datadir is None:
        locfld = os.getenv('LOCALFOLDER', default='f:/videos/meteorcam/fireballs')
        datadir, _ = os.path.split(locfld)
    if outdir is None:
        outdir = '.'
    print(f'stacking {camlist} for dates {start},{end}, reading from {datadir}, saving to {outdir}')

    sdt = datetime.datetime.strptime(start, '%Y%m%d')
    edt = datetime.datetime.strptime(end, '%Y%m%d')
    d = sdt
    dir_paths = []
    while d <= edt:
        dtstr = d.strftime('%Y%m%d')
        for cam in camlist:
            cam = cam.upper()
            camdir = os.path.join(datadir, cam, 'ConfirmedFiles')
            dirs = glob.glob1(camdir, f'{cam}_{dtstr}*')
            if len(dirs) > 0: 
                srcdir = os.path.join(camdir, dirs[0])
                ftpdet = os.path.join(srcdir, f'FTPdetectinfo_{dirs[0]}.txt')
                if os.path.isfile(ftpdet):
                    print(f'using {ftpdet}')
                    dir_paths.append(srcdir)
        d = d + datetime.timedelta(days=1)

    config = cr.loadConfigFromDirectory('.config', dir_paths[0])
    if shwr is not None:
        shwr = [shwr]

    trackStack(dir_paths, config, border=5, background_compensation=True, 
        hide_plot=noplot, showers=shwr, darkbackground=False, out_dir=outdir, scalefactor=scale,
        draw_constellations=draw_cons)

    if len(camlist) > 1:
        bn = os.path.basename(dir_paths[-1])
        outf = os.path.join(outdir, bn + "_track_stack.jpg")
        bn = f'{len(camlist)}CAMS{bn[6:]}'
        newf = os.path.join(outdir, bn + "_track_stack.jpg")
        if os.path.isfile(newf):
            os.remove(newf)
        os.rename(outf, newf)

    return


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser(description="Perform multiday/camera trackstack.")

    arg_parser.add_argument('cams', metavar='CAMS', nargs='+', type=str,
        help='List of cameras, comma-separated.')

    arg_parser.add_argument('dates', metavar='DATES', nargs='+', type=str,
        help='Start and end date, comma-separated in yyyymmdd format.')

    arg_parser.add_argument('-s', '--shower', metavar='SHOWER', type=str,
        help="Associate just this single shower given its code (e.g. PER, ORI, ETA).")

    arg_parser.add_argument('-o', '--outdir', metavar='OUTDIR', type=str,
        help="Where to save the output file.")

    arg_parser.add_argument('-f', '--scalefactor', metavar='SCALEFACTOR', type=int,
        help="Scale factor to apply when creating canvas.")

    arg_parser.add_argument('--constellations', help="Overplot constellations", action="store_true")

    cml_args = arg_parser.parse_args()
    cams = cml_args.cams[0]
    if ',' in cams:
        cams = cams.split(',')
    else:
        cams = [cams]

    dates = cml_args.dates[0]
    if ',' in dates:
        start, end = dates.split(',')
    else:
        start = dates
        end = start

    shwr = None
    if cml_args.shower:
        shwr = cml_args.shower
        if shwr == 'ALL': # special case for ALLs
            shwr = None
    
    outdir = '.'
    if cml_args.outdir:
        outdir = cml_args.outdir

    scalefactor = 1
    if cml_args.scalefactor:
        scalefactor = cml_args.scalefactor

    multiTrackStack(cams, start, end, outdir, shwr, scalefactor,draw_constellations=cml_args.constellations)



# Various maths functions scraped from WMPL

import numpy as np
import math
from datetime import datetime, timedelta, MINYEAR
# Copyright (C) 2018-2023 Mark McIntyre

# Define Julian epoch
JULIAN_EPOCH = datetime(2000, 1, 1, 12) # J2000.0 noon
J2000_JD = timedelta(2451545) # J2000.0 epoch in julian days


def jd2Date(jd, UT_corr=0, dt_obj=False):
    """ Converts the given Julian date to (year, month, day, hour, minute, second, millisecond) tuple or a datetime.  

    Arguments:  
        jd: [float] Julian date  

    Keyword arguments:  
        UT_corr: [float] UT correction in hours (difference from local time to UT)  
        dt_obj: [bool] returns a datetime object if True. False by default.  

    Return:  
        (year, month, day, hour, minute, second, millisecond) or a datetime

    """

    try:

        dt = timedelta(days=jd)
        
        date = dt + JULIAN_EPOCH - J2000_JD + timedelta(hours=UT_corr) 

    # If the date is out of range (i.e. before year 1) use year 1. This is the limitation in the datetime
    # library. Time handling should be switched to astropy.time
    except OverflowError:
        date = datetime(MINYEAR, 1, 1, 0, 0, 0)


    # Return a datetime object if dt_obj == True
    if dt_obj:
        return date

    return date.year, date.month, date.day, date.hour, date.minute, date.second, date.microsecond/1000.0


def datetime2JD(dt):
    """ Convert a datetime to a julian date  
    """

    return date2JD(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond/1000.0)


def date2JD(year, month, day, hour, minute, second, millisecond=0, UT_corr=0.0):
    """ Convert date and time to Julian Date in J2000.0.  
    
    Arguments:  
        year: [int] year  
        month: [int] month  
        day: [int] day of the date  
        hour: [int] hours  
        minute: [int] minutes  
        second: [int] seconds  

    Kwargs:  
        millisecond: [int] milliseconds (optional)  
        UT_corr: [float] UT correction in hours (difference from local time to UT)  
    
    Return:  
        [float] julian date, J2000.0 epoch  
    """

    # Convert all input arguments to integer (except milliseconds)
    year, month, day, hour, minute, second = map(int, (year, month, day, hour, minute, second))

    # Create datetime object of current time
    dt = datetime(year, month, day, hour, minute, second, int(millisecond*1000))

    # Calculate Julian date
    julian = dt - JULIAN_EPOCH + J2000_JD - timedelta(hours=UT_corr)
    
    # Convert seconds to day fractions
    return julian.days + (julian.seconds + julian.microseconds/1000000.0)/86400.0


def jd2LST(julian_date, lon):
    """ Convert Julian date to Local Sidereal Time and Greenwich Sidereal Time. The times used are apparent 
        times, not mean times.  

    Source: J. Meeus: Astronomical Algorithms  

    Arguments:  
        julian_date: [float] decimal julian date, epoch J2000.0  
        lon: [float] longitude of the observer in degrees  
    
    Return:  
        (LST, GST): [tuple of floats] a tuple of Local Sidereal Time and Greenwich Sidereal Time  
    """

    # t = (julian_date - J2000_JD.days)/36525.0

    # Greenwich Sidereal Time
    #GST = 280.46061837 + 360.98564736629*(julian_date - J2000_JD.days) + 0.000387933*t**2 - (t**3)/38710000
    #GST = (GST + 360)%360

    GST = np.degrees(calcApparentSiderealEarthRotation(julian_date))

    # Local Sidereal Time
    LST = (GST + lon + 360) % 360
    
    return LST, GST


def jd2DynamicalTimeJD(jd):
    """ Converts the given Julian date to dynamical time (i.e. Terrestrial Time, TT) Julian date. The 
        conversion takes care of leap seconds.   

    Arguments:  
        jd: [float] Julian date.  

    Return:  
        [float] Dynamical time Julian date.  
    """

    # Leap seconds as of 2017 (default)
    leap_secs = 37.0


    # Get the relevant number of leap seconds for the given JD
    # COMMENTED OUT as tai-utc.dat is empty
    #for jd_leap, ls in config.leap_seconds:
    #    
    #    if jd > jd_leap:
    #        leap_secs = ls
            

    # Calculate the dynamical JD
    jd_dyn = jd + (leap_secs + 32.184)/86400.0


    return jd_dyn


def greatCircleDistance(lat1, lon1, lat2, lon2):
    """ Calculate the great circle distance in kilometers between two points on the Earth. 
        Source: https://gis.stackexchange.com/a/56589/15183  

    Arguments:  
        lat1: [float] Latitude 1 (radians).  
        lon1: [float] Longitude 1 (radians).  
        lat2: [float] Latitude 2 (radians).  
        lon2: [float] Longitude 2 (radians).  

    Return:  
        [float]: Distance in kilometers.  
    """
    
    # Haversine formula
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 

    a = np.sin(dlat/2)**2 + np.cos(lat1)*np.cos(lat2)*np.sin(dlon/2)**2
    c = 2*np.arcsin(np.sqrt(a))

    # Distance in kilometers.
    dist = 6371*c

    return dist


def angleBetweenSphericalCoords(phi1, lambda1, phi2, lambda2):
    """ Calculates the angle between two points on a sphere.  
    
    Arguments:  
        phi1: [float] Latitude 1 (radians).  
        lambda1: [float] Longitude 1 (radians).  
        phi2: [float] Latitude 2 (radians).  
        lambda2: [float] Longitude 2 (radians).  

    Return:  
        [float] Angle between two coordinates (radians).  
    """

    return np.arccos(np.sin(phi1)*np.sin(phi2) + np.cos(phi1)*np.cos(phi2)*np.cos(lambda2 - lambda1))


def equatorialCoordPrecession(start_epoch, final_epoch, ra, dec):
    """ Precess right Ascension and declination from one epoch to another, taking only precession into 
        account.  

        Implemented from: Jean Meeus - Astronomical Algorithms, 2nd edition, pages 134-135  
    
    Arguments:  
        start_epoch: [float] Julian date of the starting epoch  
        final_epoch: [float] Julian date of the final epoch  
        ra: [float] non-corrected right ascension in radians  
        dec: [float] non-corrected declination in radians  
    
    Return:  
        (ra, dec): [tuple of floats] precessed equatorial coordinates in radians  

    """


    T = (start_epoch - J2000_JD.days)/36525.0
    t = (final_epoch - start_epoch)/36525.0

    # Calculate correction parameters
    zeta = ((2306.2181 + 1.39656*T - 0.000139*T**2)*t + (0.30188 - 0.000344*T)*t**2 + 0.017998*t**3)/3600
    z = ((2306.2181 + 1.39656*T - 0.000139*T**2)*t + (1.09468 + 0.000066*T)*t**2 + 0.018203*t**3)/3600
    theta = ((2004.3109 - 0.85330*T - 0.000217*T**2)*t - (0.42665 + 0.000217*T)*t**2 - 0.041833*t**3)/3600

    # Convert parameters to radians
    zeta, z, theta = map(math.radians, (zeta, z, theta))

    # Calculate the next set of parameters
    A = np.cos(dec) * np.sin(ra + zeta)
    B = np.cos(theta)* np.cos(dec) * np.cos(ra + zeta) - np.sin(theta) * np.sin(dec)
    C = np.sin(theta)* np.cos(dec) * np.cos(ra + zeta) + np.cos(theta) * np.sin(dec)

    # Calculate right ascension
    ra_corr = np.arctan2(A, B) + z

    # Calculate declination (apply a different equation if close to the pole, closer then 0.5 degrees)
    if (np.pi/2 - np.abs(dec)) < np.radians(0.5):
        dec_corr = np.sign(C)*np.arccos(np.sqrt(A**2 + B**2))
    else:
        dec_corr = np.arcsin(C)

    # Wrap right ascension to [0, 2*pi] range
    ra_corr = ra_corr % (2*np.pi)

    # Wrap declination to [-pi/2, pi/2] range
    dec_corr = (dec_corr + np.pi/2) % np.pi - np.pi/2

    return ra_corr, dec_corr


equatorialCoordPrecession_vect = np.vectorize(equatorialCoordPrecession, excluded=['start_epoch'])
"""Vectorize the equatorialCoordPrecession, so ra and dec can be passed as numpy arrays"""


def raDec2AltAz(ra, dec, jd, lat, lon):
    """ Convert right ascension and declination to azimuth (+east of sue north) and altitude.  

    Arguments:  
        ra: [float] right ascension in radians  
        dec: [float] declination in radians  
        jd: [float] Julian date  
        lat: [float] latitude in radians  
        lon: [float] longitude in radians  

    Return:  
        (azim, elev): [tuple]  
            azim: [float] azimuth (+east of due north) in radians  
            elev: [float] elevation above horizon in radians  

        """

    # Calculate Local Sidereal Time
    lst = np.radians(jd2LST(jd, np.degrees(lon))[0])

    # Calculate the hour angle
    ha = lst - ra

    # Constrain the hour angle to [-pi, pi] range
    ha = (ha + np.pi) % (2*np.pi) - np.pi

    # Calculate the azimuth
    azim = np.pi + np.arctan2(np.sin(ha), np.cos(ha)*np.sin(lat) - np.tan(dec)*np.cos(lat))

    # Calculate the sine of elevation
    sin_elev = np.sin(lat)*np.sin(dec) + np.cos(lat)*np.cos(dec)*np.cos(ha)

    # Wrap the sine of elevation in the [-1, +1] range
    sin_elev = (sin_elev + 1) % 2 - 1

    elev = np.arcsin(sin_elev)

    return azim, elev


raDec2AltAz_vect = np.vectorize(raDec2AltAz, excluded=['lat', 'lon'])
"""Vectorize the raDec2AltAz function so it can take numpy arrays for: ra, dec, jd"""


def altAz2RADec(azim, elev, jd, lat, lon):
    """ Convert azimuth and altitude in a given time and position on Earth to right ascension and 
        declination.   

    Arguments:  
        azim: [float] azimuth (+east of due north) in radians  
        elev: [float] elevation above horizon in radians  
        jd: [float] Julian date  
        lat: [float] latitude of the observer in radians  
        lon: [float] longitde of the observer in radians  

    Return:  
        (RA, dec): [tuple]  
            RA: [float] right ascension (radians)  
            dec: [float] declination (radians)  
    """
    
    # Calculate hour angle
    ha = np.arctan2(-np.sin(azim), np.tan(elev)*np.cos(lat) - np.cos(azim)*np.sin(lat))

    # Calculate Local Sidereal Time
    lst = np.radians(jd2LST(jd, np.degrees(lon))[0])
    
    # Calculate right ascension
    ra = (lst - ha) % (2*np.pi)

    # Calculate declination
    dec = np.arcsin(np.sin(lat)*np.sin(elev) + np.cos(lat)*np.cos(elev)*np.cos(azim))

    return ra, dec


altAz2RADec_vect = np.vectorize(altAz2RADec, excluded=['lat', 'lon'])
""" Vectorize the altAz2RADec function so it can take numpy arrays for: azim, elev, jd """


def calcApparentSiderealEarthRotation(julian_date):
    """ Calculate apparent sidereal rotation GST of the Earth.  
        
        Calculated according to:  
        Clark, D. L. (2010). Searching for fireball pre-detections in sky surveys. The School of Graduate and 
        Postdoctoral Studies. University of Western Ontario, London, Ontario, Canada, MSc Thesis.  

    """

    t = (julian_date - J2000_JD.days)/36525.0

    # Calculate the Mean sidereal rotation of the Earth in radians (Greenwich Sidereal Time)
    GST = 280.46061837 + 360.98564736629*(julian_date - J2000_JD.days) + 0.000387933*t**2 - (t**3)/38710000
    GST = (GST + 360) % 360
    GST = math.radians(GST)

    # print('GST:', np.degrees(GST), 'deg')

    # Calculate the dynamical time JD
    jd_dyn = jd2DynamicalTimeJD(julian_date)


    # Calculate Earth's nutation components
    delta_psi, delta_eps = calcNutationComponents(jd_dyn)

    # print('Delta Psi:', np.degrees(delta_psi), 'deg')
    # print('Delta Epsilon:', np.degrees(delta_eps), 'deg')


    # Calculate the mean obliquity (in arcsec)
    u = (jd_dyn - 2451545.0)/3652500.0
    eps0 = 84381.448 - 4680.93*u - 1.55*u**2 + 1999.25*u**3 - 51.38*u**4 - 249.67*u**5 - 39.05*u**6 \
        + 7.12*u**7 + 27.87*u**8 + 5.79*u**9 + 2.45*u**10

    # Convert to radians
    eps0 /= 3600
    eps0 = np.radians(eps0)

    # print('Mean obliquity:', np.degrees(eps0), 'deg')

    # Calculate apparent sidereal Earth's rotation
    app_sid_rot = (GST + delta_psi*math.cos(eps0 + delta_eps)) % (2*math.pi)

    return app_sid_rot


def calcNutationComponents(jd_dyn):
    """ Calculate Earth's nutation components from the given Julian date.  

    Source: Meeus (1998) Astronomical algorithms, 2nd edition, chapter 22.  
    
    The precision is limited to 0.5" in nutation in longitude and 0.1" in nutation in obliquity. The errata 
    for the 2nd edition was used to correct the equation for delta_psi.
    
    Arguments:  
        jd_dyn: [float] Dynamical Julian date. See wmpl.Utils.TrajConversions.jd2DynamicalTimeJD function.  

    Return:  
        (delta_psi, delta_eps): [tuple of floats] Differences from mean nutation due to the influence of
            the Moon and minor effects (radians).  
    """


    T = (jd_dyn - J2000_JD.days)/36525.0

    # # Mean Elongation of the Moon from the Sun
    # D = 297.85036 + 445267.11148*T - 0.0019142*T**2 + (T**3)/189474

    # # Mean anomaly of the Earth with respect to the Sun
    # M = 357.52772 + 35999.05034*T - 0.0001603*T**2 - (T**3)/300000

    # # Mean anomaly of the Moon
    # Mm = 134.96298 + 477198.867398*T + 0.0086972*T**2 + (T**3)/56250

    # # Argument of latitude of the Moon
    # F = 93.27191  + 483202.017538*T - 0.0036825*T**2 + (T**3)/327270


    # Longitude of the ascending node of the Moon's mean orbit on the ecliptic, measured from the mean equinox
    # of the date
    omega = 125.04452 - 1934.136261*T


    # Mean longitude of the Sun (deg)
    L = 280.4665 + 36000.7698*T

    # Mean longitude of the Moon (deg)
    Ll = 218.3165 + 481267.8813*T


    # Nutation in longitude
    delta_psi = -17.2*math.sin(math.radians(omega)) - 1.32*math.sin(np.radians(2*L)) \
        - 0.23*math.sin(math.radians(2*Ll)) + 0.21*math.sin(math.radians(2*omega))

    # Nutation in obliquity
    delta_eps = 9.2*math.cos(math.radians(omega)) + 0.57*math.cos(math.radians(2*L)) \
        + 0.1*math.cos(math.radians(2*Ll)) - 0.09*math.cos(math.radians(2*omega))


    # Convert to radians
    delta_psi = np.radians(delta_psi/3600)
    delta_eps = np.radians(delta_eps/3600)

    return delta_psi, delta_eps



# Copyright (C) 2018-2023 Mark McIntyre

import requests
import os


def getECSVs(stationID, dateStr, savefiles=False, outdir='.'):
    """
    Retrieve a detection in ECSV format for the specified date  

    Arguments:  
        stationID:  [str] RMS Station ID code  
        dateStr:    [str] Date/time to retrieve for in ISO1601 format   
                          eg 2021-07-17T02:41:05.05  
    
    Keyword Arguments:  
        saveFiles:  [bool] save to file, or print to screen. Default False  
        outdir:     [str] path to save files into. Default '.'  
    """
    apiurl='https://api.ukmeteornetwork.co.uk/getecsv?stat={}&dt={}'
    res = requests.get(apiurl.format(stationID, dateStr))
    ecsvlines=''
    if res.status_code == 200:
        rawdata=res.text.strip()
        if len(rawdata) > 10:
            ecsvlines=rawdata.split('\n') # convert the raw data into a python list
            if savefiles is True:
                numecsvs = len([e for e in ecsvlines if '# %ECSV' in e]) # find out how many meteors 
                fnamebase = dateStr.replace(':','_').replace('.','_') # create an output filename
                if numecsvs == 1:
                    print('saving to ', fnamebase+ '.ecsv')
                    with open(os.path.join(outdir, fnamebase + '.ecsv'), 'w') as outf:
                        outf.writelines(ecsvlines)
                else:
                    outf = None
                    j=1
                    for i in range(len(ecsvlines)):
                        if '# %ECSV' in ecsvlines[i]:
                            if outf is not None:
                                outf.close()
                                j=j+1
                            outf = open(os.path.join(outdir, fnamebase + f'_M{j:03d}.ecsv'), 'w')
                            print('saving to ', fnamebase + f'_M{j:03d}.ecsv')
                        outf.write(f'{ecsvlines[i]}\n')
        else:
            print('no error, but no data returned')
    else:
        print(res.status_code)
    return ecsvlines



# Copyright (C) 2018-2023 Mark McIntyre
import os
import sys
import platform
import base64
from email.mime.text import MIMEText

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

# If modifying these scopes, delete the file token.json.
SCOPES =['https://mail.google.com/']


def _getGmailCreds():
    creds = None
    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    tokfile = os.path.expanduser('~/.ssh/gmailtoken.json')
    crdfile = os.path.expanduser('~/.ssh/gmailcreds.json')
    if os.path.exists(tokfile):
        creds = Credentials.from_authorized_user_file(tokfile, SCOPES)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            if not os.path.isfile(crdfile):
                print('to use this function you must have stored your Google OAUTH2 secret json file in ~/.ssh/gmailcreds.json.')
                print('To set up OAUTH2 go to your google cloud console, select APIs then Credentials, and add an OAUTH2 desktop client ID.')
                return None
            flow = InstalledAppFlow.from_client_secrets_file(crdfile, SCOPES)
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(tokfile, 'w') as token:
            token.write(creds.to_json())
    return creds


def _create_message(sender, to, subject, message_text):
    msg = MIMEText(message_text)
    msg['to'] = to
    msg['from'] = sender
    msg['subject'] = subject
    return {'raw': base64.urlsafe_b64encode(msg.as_string().encode('utf-8')).decode('utf-8')}



def sendAnEmail(mailrecip, message, msgtype, mailfrom, files=None):
    """ sends an email using gmail. 
    
        Arguments:  
            mailrecip:  [string] email address of recipient.  
            message:    [string] the message to send.  
            msgtype:    [string] Prefix for the subject line, eg Test, Warning.  
            mailfrom:   [string] email address of sender.   
            files:      [list]   list of files to attach, not currently implemented.  

        Returns:  
            Nothing, though a message is printed onscreen.  

        Notes:  
            You must have gmail OAUTH2 set up. The gmail credentials must be stored as follows:  
                token: $HOME/.ssh/gmailtoken.json  
                creds: $HOME/.ssh/gmailcreds.json  

            On Windows, $HOME corresponds to c:/users/yourid. If there is no .ssh folder, create it.   
        """
    
    if msgtype is None:
        msgtype = platform.uname()[1]

    # email a summary to the mailrecip
    creds = _getGmailCreds()
    if not creds:
        return 
    service = build('gmail', 'v1', credentials=creds)

    subj ='{:s}: {:s}'.format(msgtype, message[:30])
    message = '{:s}: {:s}'.format(msgtype, message)

    mailmsg = _create_message(mailfrom, mailrecip, subj, message)

    try:
        retval = (service.users().messages().send(userId='me', body=mailmsg).execute())
        print('Message Id: %s' % retval['id'])
    except:
        print('An error occurred sending the message')


if __name__ == '__main__':
    if len(sys.argv) < 4:
        print('usage: sendAnEmail.py recipient "message in quotes" Alert|Warning|Error sender ')
    else:
        sendAnEmail(sys.argv[1], sys.argv[2], sys.argv[3],sys.argv[4])



#
# python module to read the IMO Working Shower short List
#
# Copyright (C) 2018-2023 Mark McIntyre

import xmltodict
import datetime
import os
import numpy as np
from ukmon_meteortools.utils import jd2Date, sollon2jd

# imported from $SRC/share
try:
    from majorminor import majorlist, minorlist
except:
    majorlist = ['QUA', 'LYR', 'ETA', 'CAP', 'SDA', 'PER', 'AUR', 'ORI', 'NTA', 'STA', 'LEO', 'GEM', 'URS']
    minorlist = ['SPE','OCT','DRA','EGE','MON','HYD','COM','NOO']


class IMOshowerList:
    """
    Class that loads and parses the IMO Working Shower list, or if needed, the unconfirmed list. 
    Not all known showers are in the IMO working list. If a shower is not in the Working List then 
    this library will reference the full shower list curated by Peter Jenniskens which contains 
    debated and unconfirmed showers. 

    These list are updated whenever the library version is bumped, but if you want to override the files, define an 
    environment variable DATADIR, and place your own copies of the files at $DATADIR/share. See the share submodule 
    for more information. 

    """
    def __init__(self, fname=None, fullstreamname=None):
        if fname is None:
            datadir = os.getenv('DATADIR', default='/home/ec2-user/prod/data')
            fname = os.path.join(datadir, '..', 'share', 'IMO_Working_Meteor_Shower_List.xml')
            if not os.path.isfile(fname):
                datadir=os.path.split(os.path.abspath(__file__))[0]
                fname = os.path.join(datadir, '..', 'share', 'IMO_Working_Meteor_Shower_List.xml')
        
        tmplist = xmltodict.parse(open(fname, 'rb').read())
        self.showerlist = tmplist['meteor_shower_list']['shower']
        self.fullstreamdata = None

    
    def getShowerByCode(self, iaucode):
        ds = {'@id':None, 'IAU_code':None,'start':None, 'end':None, 
            'peak':None, 'r':None, 'name':None, 'V':None, 'ZHR':None, 'RA':None, 'DE':None}
        got = None
        for shower in self.showerlist:
            if shower['IAU_code'] == iaucode:
                return shower
        if got is None:
            # load stream full dataset as backup
            if self.fullstreamdata is None:
                wmpldir = os.getenv('WMPL_LOC', default='/home/ec2-user/src/WesternMeteorPyLib')
                fullstreamname = os.path.join(wmpldir, 'wmpl', 'share', 'streamfulldata.npy')
                if not os.path.isfile(fullstreamname):
                    datadir=os.path.split(os.path.abspath(__file__))[0]
                    fullstreamname = os.path.join(datadir, '..', 'share', 'streamfulldata.npy')
                self.fullstreamdata = np.load(fullstreamname)
            subset = self.fullstreamdata[np.where(self.fullstreamdata[:,3]==iaucode)]
            
            if subset is not None:
                mtch = [sh for sh in subset if sh[6] != '-2']
                if len(mtch) > 0:
                    ds['IAU_code'] = mtch[-1][3].strip()
                    ds['name'] = mtch[-1][4].strip()
                    ds['V'] = mtch[-1][12]
                    ds['@id'] = mtch[-1][1]
                    ds['RA'] = mtch[-1][8]
                    ds['DE'] = mtch[-1][9]

                    pksollong = float(mtch[-1][7])
                    dt = datetime.datetime.now()
                    yr = dt.year
                    mth = dt.month
                    jd = sollon2jd(yr, mth, pksollong)
                    pkdt = jd2Date(jd, dt_obj=True)
                    ds['peak'] = pkdt.strftime('%h %d')
                    # start/end pop idx, ZHR not available in the IAU data
                    ds['start'] = (pkdt + datetime.timedelta(days=-2)).strftime('%h %d')
                    ds['end'] = (pkdt + datetime.timedelta(days=2)).strftime('%h %d')
                    ds['r'] = '2.0'
                    ds['ZHR'] = '3.0' 
        return ds

    def getStart(self, iaucode, currdt=None):
        shower = self.getShowerByCode(iaucode)
        if currdt is None:
            now = datetime.datetime.today().year
            mth = datetime.datetime.today().month
        else:
            now = datetime.datetime.strptime(str(currdt), '%Y%m%d')
            mth = now.month
            now = now.year
        startdate = datetime.datetime.strptime(shower['start'], '%b %d')
        if iaucode == 'QUA' and mth !=12:
            # quadrantids straddle yearend
            now = now - 1
        startdate = startdate.replace(year=now)
        return startdate

    def getEnd(self, iaucode, currdt=None):
        shower = self.getShowerByCode(iaucode)
        if currdt is None:
            now = datetime.datetime.today().year
            mth = datetime.datetime.today().month
        else:
            now = datetime.datetime.strptime(str(currdt), '%Y%m%d')
            mth = now.month
            now = now.year
        enddate = datetime.datetime.strptime(shower['end'], '%b %d')
        if iaucode == 'QUA' and mth == 12:
            # quadrantids straddle yearend
            now = now + 1
        enddate = enddate.replace(year=now)
        return enddate

    def getPeak(self, iaucode, currdt=None):
        shower = self.getShowerByCode(iaucode)
        if currdt is None:
            now = datetime.datetime.today().year
            mth = datetime.datetime.today().month
        else:
            now = datetime.datetime.strptime(str(currdt), '%Y%m%d')
            mth = now.month
            now = now.year
        enddate = datetime.datetime.strptime(shower['peak'], '%b %d')
        if iaucode == 'QUA' and mth == 12:
            # quadrantids straddle yearend
            now = now + 1
        enddate = enddate.replace(year=now)
        return enddate

    def getRvalue(self, iaucode):
        shower = self.getShowerByCode(iaucode)
        return shower['r']

    def getName(self, iaucode):
        shower = self.getShowerByCode(iaucode)
        return shower['name']

    def getVelocity(self, iaucode):
        shower = self.getShowerByCode(iaucode)
        return shower['V']

    def getZHR(self, iaucode):
        shower = self.getShowerByCode(iaucode)
        zhr = shower['ZHR']
        if zhr is None:
            return -1
        else:
            return int(zhr)

    def getRaDec(self, iaucode):
        shower = self.getShowerByCode(iaucode)
        return float(shower['RA']), float(shower['DE'])

    def getActiveShowers(self, datetotest, majorOnly=False, inclMinor=False):
        activelist = []
        for shower in self.showerlist:
            shwname = shower['IAU_code']
            start = self.getStart(shwname, datetotest.strftime('%Y%m%d'))
            end = self.getEnd(shwname, datetotest.strftime('%Y%m%d')) + datetime.timedelta(days=3)
            if datetotest > start and datetotest < end:
                if majorOnly is False or (majorOnly is True and shwname in majorlist):
                    activelist.append(shwname)
                elif inclMinor is True and shwname in minorlist:
                    activelist.append(shwname)
        return activelist

    def getMajorShowers(self, includeSpo=False, stringFmt=False):
        majlist = majorlist 
        if includeSpo is True:
            majlist.append('spo')
        if stringFmt is True:
            tmplist = ''
            for shwr in majlist:
                tmplist = tmplist + shwr + ' '
            majlist = tmplist
        return majlist



# Python class to parse a UFOCapture XML files
#
# disable some linter warnings
# Copyright (C) 2018-2023 Mark McIntyre

import xmltodict
import numpy


class UCXml:
    """ Load and manage a UFOCapture XML file containing potential detections """
    MAXGAP = 50

    def setMaxGap(self, mg):
        self.MAXGAP = mg

    def __init__(self, fname):
        """ Construct a UCXML object from fname """
        with open(fname) as fd:
            self.ucxml = xmltodict.parse(fd.read())

    def getDate(self):
        ur = self.ucxml['ufocapture_record']
        return int(ur['@y']) * 10000 + int(ur['@mo']) * 100 + int(ur['@d'])

    def getDateYMD(self):
        yr = int(self.ucxml['ufocapture_record']['@y'])
        mo = int(self.ucxml['ufocapture_record']['@mo'])
        dy = int(self.ucxml['ufocapture_record']['@d'])
        return yr, mo, dy

    def getDateStr(self):
        ur = self.ucxml['ufocapture_record']
        return ur['@y'] + "-" + ur['@mo'] + "-" + ur['@d']

    def getTime(self):
        ur = self.ucxml['ufocapture_record']
        return int(ur['@h']) * 3600 + int(ur['@m']) * 60 + float(ur['@s'])

    def getTimeStr(self):
        ur = self.ucxml['ufocapture_record']
        return ur['@h'] + ":" + ur['@m'] + ":" + ur['@s']

    def getTimeHMS(self):
        h = int(self.ucxml['ufocapture_record']['@h'])
        m = int(self.ucxml['ufocapture_record']['@m'])
        s = float(self.ucxml['ufocapture_record']['@s'])
        return h, m, s

    def getCameraDetails(self):
        fps = float(self.ucxml['ufocapture_record']['@fps'])
        cx = float(self.ucxml['ufocapture_record']['@cx'])
        cy = float(self.ucxml['ufocapture_record']['@cy'])
        return fps, cx, cy

    def getStationDetails(self):
        sta = self.ucxml['ufocapture_record']['@lid'] + "_" + self.ucxml['ufocapture_record']['@sid']
        lid = self.ucxml['ufocapture_record']['@lid']
        sid = self.ucxml['ufocapture_record']['@sid']
        lat = float(self.ucxml['ufocapture_record']['@lat'])
        lng = float(self.ucxml['ufocapture_record']['@lng'])
        alt = float(self.ucxml['ufocapture_record']['@alt'])
        return sta, lid, sid, lat, lng, alt

    def getHits(self):
        uc = self.ucxml['ufocapture_record']['ufocapture_paths']
        return int(uc['@hit'])

    def getNumObjs(self):
        try:
            uc = self.ucxml['ufocapture_record']['ufocapture_paths']
        except:
            return 0, None
        nhits = int(uc['@hit'])
        objlist = numpy.zeros(nhits)
        nobjs = 0
        for i in range(nhits):
            if nhits == 1:
                p = uc['uc_path']
            else:
                p = uc['uc_path'][i]
            ono = int(p['@ono'])
            if ono not in objlist:
                objlist[nobjs] = ono
                nobjs = nobjs + 1
        objlist = numpy.resize(objlist, nobjs)
        return nobjs, objlist

    def getPath(self):
        try:
            uc = self.ucxml['ufocapture_record']['ufocapture_paths']
        except:
            # print('xml file not valid - no ufocapture_paths')
            pathx = numpy.zeros(1)
            pathy = numpy.zeros(1)
            bri = numpy.zeros(1)
            return pathx, pathy, bri, 0
        nhits = int(uc['@hit'])
        pathx = numpy.zeros(nhits)
        pathy = numpy.zeros(nhits)
        bri = numpy.zeros(nhits)
        if nhits < 2:
            return pathx, pathy, bri, 0
        lastono = 0
        for i in range(nhits):
            p = uc['uc_path'][i]
            ono = int(p['@ono'])
            # its possible for one path to include two events
            # at different times. This is a UFO feature...
            if lastono > 0 and lastono != ono:
                pathx = numpy.resize(pathx, i)
                pathy = numpy.resize(pathy, i)
                bri = numpy.resize(bri, i)
                return pathx, pathy, bri, i
            lastono = ono
            pathx[i] = p['@x']
            pathy[i] = p['@y']
            bri[i] = p['@bmax']
        return pathx, pathy, bri, 0

    def getPathv2(self, objno):
        try:
            uc = self.ucxml['ufocapture_record']['ufocapture_paths']
        except:
            pathx = numpy.zeros(1)
            pathy = numpy.zeros(1)
            bri = numpy.zeros(1)
            pxls = numpy.zeros(1)
            fnos = numpy.zeros(1)
            return pathx, pathy, bri, pxls, fnos
        pfno = 0  # previous frame number, to check for unrealistic gaps in trails
        fno = 0  # current frame number, to check for unrealistic gaps in trails
        nhits = int(uc['@hit'])
        pathx = numpy.zeros(nhits)
        pathy = numpy.zeros(nhits)
        bri = numpy.zeros(nhits)
        pxls = numpy.zeros(nhits)
        fnos = numpy.zeros(nhits)
        j = 0
        for i in range(nhits):
            if nhits == 1:
                p = uc['uc_path']
            else:
                p = uc['uc_path'][i]
            ono = int(p['@ono'])
            # npx = int(p['@pixel'])
            fno = int(p['@fno'])
            if pfno == 0:
                pfno = fno
            #fdiff = fno - pfno
            if ono == objno:  # and fdiff < self.MAXGAP:  # and npx > 3:
                # its possible for one path to include two events
                # at different times. This is a UFO feature...
                pathx[j] = p['@x']
                pathy[j] = p['@y']
                bri[j] = p['@bmax']
                pxls[j] = p['@pixel']
                fnos[j] = p['@fno']
                j = j + 1
        if j > 0:
            pathx = numpy.resize(pathx, j)
            pathy = numpy.resize(pathy, j)
            bri = numpy.resize(bri, j)
            pxls = numpy.resize(pxls, j)
            fnos = numpy.resize(fnos, j)
        else:
            pathx = numpy.zeros(1)
            pathy = numpy.zeros(1)
            bri = numpy.zeros(1)
            pxls = numpy.zeros(1)
            fnos = numpy.zeros(1)
        return pathx, pathy, bri, pxls, fnos

    def getPathElement(self, fno):
        uc = self.ucxml['ufocapture_record']['ufocapture_paths']
        pth = uc['uc_path'][fno]
        return pth['@fno'], pth['@ono'], pth['@pixel'], pth['@bmax'], pth['@x'], pth['@y']


if __name__ == '__main__':
    dd = UCXml('test_data/M20200427_201548_TACKLEY_NE.XML')

    y, m, d = dd.getDateYMD()
    h, mi, s = dd.getTimeHMS()
    print('date and time', y, m, d, h, mi, s)

    station, lid, sid, lat, lng, alti = dd.getStationDetails()
    fps, cx, cy = dd.getCameraDetails()
    print('location', station, lat, lng, alti)
    print('camera', fps, cx, cy)

    nhits = dd.getHits()
    for i in range(nhits):
        fno, ono, pixel, bmax, x, y = dd.getPathElement(i)
        print(fno, pixel, bmax, x, y)



# Copyright (C) 2018-2023 Mark McIntyre
#
import numpy as np
from ukmon_meteortools.utils import date2JD


def sollon2jd(Year, Month, Long):
    """
    Calculate the julian date corresponding to a solar longitude. 
    Because Solar Longitude is relative to the Spring equinox, the exact date 
    of a given LS varies from year to year.  

    Parameters:  
        Year: [int] year you wish to calculate in.  
        Month: [int] month you wish to calculate in.  
        Long:  [float] The solar longitude to convert.   

    Returns:  
        [float] julian date  

    Notes:  
        The function is only stable for date ranges from 1900-2100.  
    """

    Long = np.radians(Long)
    N = Year - 2000
    if abs(N) > 100:
        print("Algorithm is not stable for years below 1900 or above 2100")

    JDM0 = 2451182.24736 + 365.25963575 * N
    ApproxJD = date2JD(Year, Month, 15, 12, 0, 0)
    DiffJD = ApproxJD-2451545

    Dt = 1.94330 * np.sin(Long - 1.798135) + 0.01305272 * np.sin(2*Long + 2.634232) + 78.195268 + 58.13165 * Long - 0.0000089408 * DiffJD

    if abs(ApproxJD - (JDM0 + Dt))>50:
        Dt = Dt + 365.2596

    JD1 = JDM0 + Dt

    return JD1


if __name__ == '__main__':
    print(sollon2jd(2021, 1, 140))
    print(sollon2jd(2021, 8, 140))
    print(sollon2jd(2021, 12, 140))



# Copyright (C) 2018-2023 Mark McIntyre
#
# analyse a UFO dataset using RMS - well, a start at it anyway ! 
#
import os
import argparse
try: 
    import RMS.ConfigReader as cr
    from RMS.DetectStarsAndMeteors import detectStarsAndMeteors
except:
    print('RMS not available')


def analyseUFOwithRMS(config, ff_directory, ff_name):
    """
    Analyse a UFO video clip using RMS, to get a list of stars and meteors  

    Arguments:
        config: [string] An RMS-style config file for the  UFO camera  
        ff_directory: [string] The location of the file to analyse  
        ff_name: [string] A video file to analyse.   

    Returns:  
        (star_list, meteor_list): tuple containing a list of stars and a list of meteors  
        
    Note: the name of the video clip must be in %Y%m%d_%H%M%S.%f format.   
    """ 
    flat_struct=None
    dark=None
    mask=None
    _, star_list, meteor_list = detectStarsAndMeteors(ff_directory, ff_name, config, flat_struct, dark, mask)

    return star_list, meteor_list

   
if __name__ == '__main__':
    # COMMAND LINE ARGUMENTS
    # Init the command line arguments parser
    arg_parser = argparse.ArgumentParser(description="Analyses a UFO avi with RMS.")

    arg_parser.add_argument('ff_path', nargs='+', metavar='FILE_PATH', type=str,
        help='Full path and name of the file to analyse')

    arg_parser.add_argument('-c', '--config', nargs=1, metavar='CONFIG_PATH', type=str,
        help="Path to a config file which will be used instead of the default one.")
    
    # Parse the command line arguments
    cml_args = arg_parser.parse_args()

    #########################

    ff_directory, ff_name = os.path.split(cml_args.ff_path[0])

    print(ff_directory, ff_name)
    config = cr.loadConfigFromDirectory(cml_args.config[0], ff_directory)
    print("loaded config \n\n\n")
    
    analyseUFOwithRMS(config, ff_directory, ff_name)



# Copyright (C) 2018-2023 Mark McIntyre
#
# python script to get all live JPGs for a specified time
#
import os
import pandas as pd
import requests


def getLiveJpgs(dtstr, outdir=None, create_txt=False):
    """
    Retrieve live images from the ukmon website that match a pattern  

    Arguments:  
        dtstr:      [str] Date in YYYYMMDD_HHMMSS format. Partial strings allowed  
        outdir:     [str] Where to save the file. Default is to create a folder named dtstr  
        create_txt: [bool] If true, create a text file containing the pattern matches  

    Notes:  
        We only keep the last few thousand live images so this function will return nothing
        for older data. 
    """
    if outdir is None:
        outdir = dtstr
    os.makedirs(outdir, exist_ok=True)

    apiurl = 'https://api.ukmeteornetwork.co.uk/liveimages/getlive'
    liveimgs = pd.read_json(f'{apiurl}?pattern={dtstr}')

    weburl = 'https://live.ukmeteornetwork.co.uk/'

    for _, img in liveimgs.iterrows():
        try:
            jpgurl = f'{weburl}{img.image_name}'
            _download(jpgurl, outdir)
            xmlurl = jpgurl.replace('P.jpg', '.xml')
            _download(xmlurl, outdir)
            print(f'retrieved {jpgurl}')
            if create_txt:
                createTxtFile(img.image_name, outdir)
        except:
            print(f'{img.image_name} unavailable')


def getFBfiles(patt, outdir=None):
    """
    Retrieve fireball files from the ukmon website that match a pattern  

    Arguments:  
        patt:      [str] pattern to match.  
        outdir:     [str] Where to save the files. See notes.  

    Returns:  
        a pandas dataframe containing the filenames and presigned URLs

    Example:  
        import pandas as pd  
        df = getFBfiles('UK0006_20230421_2122', 'c:/temp/uk0006')  

    Notes:  
        The function retrieves the FF and FR files matching the pattern, plus the config and platepar 
        for the camera, provided the files are available on our server.  
        If outdir is not supplied, a folder will be created in the current working directory named
        using the station ID code. 
    """
    if outdir is None:
        outdir = patt[:6]
    os.makedirs(outdir, exist_ok=True)
    apiurl = 'https://api.ukmeteornetwork.co.uk/fireballs/getfb'
    fbfiles = pd.read_json(f'{apiurl}?pattern={patt}')
    if len(fbfiles) == 0:
        print('no matching data found')
        return None
    for _,fil in fbfiles.iterrows():
        fname = fil['filename']
        url = fil['url']
        _download(url, outdir, fname)
        print(fname)
    return fbfiles


def createTxtFile(fname, outdir=None):
    """
    Create a text file named after the cameraID, containing a list of fireball files 
    to be retrieved from a remote camera.  

    Arguments:  
        fname:  [str] the name of the FF file to be retrieved. 
        outdir: [str] where to save the files. See notes.  

    Notes:  
        the fname parameter should be the name of the live JPG for which you wish to 
        retrieve the corresponding FF and FR files.  
        If outdir is not supplied, the files will be saved in the current directory.   
    """
    if fname[0] == 'M':
        spls = fname.split('_')
        stationid = spls[-1][:6].lower()
        dtime = fname[1:16]
        patt = f'FF_{stationid}_{dtime}'
        stationid = stationid.lower()
    else:
        patt = fname[:25]
        stationid = fname[3:9].lower()
    if outdir is None:
        outdir = '.' 
    os.makedirs(outdir, exist_ok=True)
    
    txtf = os.path.join(outdir, f'{stationid}.txt')
    if os.path.isfile(txtf):
        os.remove(txtf)
    patt = patt.upper()
    with open(txtf,'w') as outf:
        outf.write(f'{patt}\n{patt.replace("FF_", "FR_")}\n')
    return txtf


def _download(url, outdir, fname=None):
    get_response = requests.get(url, stream=True)
    if fname is None:
        fname = url.split("/")[-1]
    with open(os.path.join(outdir, fname), 'wb') as f:
        for chunk in get_response.iter_content(chunk_size=4096):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)



#  A class to handle UFO Analyser xml files
# Copyright (C) 2018-2023 Mark McIntyre

import xmltodict
import datetime
import math
import numpy
from ukmon_meteortools.utils import datetime2JD, altAz2RADec


class UAXml:
    """" Load and manage a UFO-Analyser XML file containing meteor detections """
    def __init__(self, fname):
        """Construct a UAxml object from fname"""
        with open(fname) as fd:
            self.uaxml = xmltodict.parse(fd.read())

    def getDateTime(self):
        y, m, d = UAXml.getDateYMD(self)
        h, mi, s = UAXml.getTimeHMS(self)
        ss = int(math.floor(s))
        ms = int((s - ss) * 1000000)

        dtim = datetime.datetime(y, m, d, h, mi, ss, ms)
        return dtim

    def getDate(self):
        ur = self.uaxml['ufoanalyzer_record']
        return int(ur['@y']) * 10000 + int(ur['@mo']) * 100 + int(ur['@d'])

    def getDateYMD(self):
        yr = int(self.uaxml['ufoanalyzer_record']['@y'])
        mo = int(self.uaxml['ufoanalyzer_record']['@mo'])
        dy = int(self.uaxml['ufoanalyzer_record']['@d'])
        return yr, mo, dy

    def getDateStr(self):
        ur = self.uaxml['ufoanalyzer_record']
        return ur['@y'] + "-" + ur['@mo'] + "-" + ur['@d']

    def getTime(self):
        ur = self.uaxml['ufoanalyzer_record']
        return int(ur['@h']) * 3600 + int(ur['@m']) * 60 + float(ur['@s'])

    def getTimeStr(self):
        ur = self.uaxml['ufoanalyzer_record']
        return ur['@h'] + ":" + ur['@m'] + ":" + ur['@s']

    def getTimeHMS(self):
        h = int(self.uaxml['ufoanalyzer_record']['@h'])
        m = int(self.uaxml['ufoanalyzer_record']['@m'])
        s = float(self.uaxml['ufoanalyzer_record']['@s'])
        return h, m, s

    def getCameraDetails(self):
        fps = float(self.uaxml['ufoanalyzer_record']['@fps'])
        cx = float(self.uaxml['ufoanalyzer_record']['@cx'])
        cy = float(self.uaxml['ufoanalyzer_record']['@cy'])
        isintl = int(self.uaxml['ufoanalyzer_record']['@interlaced'])
        return fps, cx, cy, isintl

    def getStationDetails(self):
        sta = self.uaxml['ufoanalyzer_record']['@lid'] + "_" + self.uaxml['ufoanalyzer_record']['@sid']
        lid = self.uaxml['ufoanalyzer_record']['@lid']
        sid = self.uaxml['ufoanalyzer_record']['@sid']
        lat = float(self.uaxml['ufoanalyzer_record']['@lat'])
        lng = float(self.uaxml['ufoanalyzer_record']['@lng'])
        alt = float(self.uaxml['ufoanalyzer_record']['@alt'])
        return sta, lid, sid, lat, lng, alt

    def getProfDetails(self):
        az = float(self.uaxml['ufoanalyzer_record']['@az'])
        ev = float(self.uaxml['ufoanalyzer_record']['@ev'])
        rot = float(self.uaxml['ufoanalyzer_record']['@rot'])
        fovh = float(self.uaxml['ufoanalyzer_record']['@vx'])
        yx = float(self.uaxml['ufoanalyzer_record']['@yx'])
        dx = float(self.uaxml['ufoanalyzer_record']['@dx'])
        dy = float(self.uaxml['ufoanalyzer_record']['@dy'])
        lnk = float(self.uaxml['ufoanalyzer_record']['@dl'])
        return az, ev, rot, fovh, yx, dx, dy, lnk

    def getObjectCount(self):
        return int(self.uaxml['ufoanalyzer_record']['@o'])

    def getObjectBasics(self, objno):
        uos = self.uaxml['ufoanalyzer_record']['ua2_objects']
        oc = int(self.uaxml['ufoanalyzer_record']['@o'])
        if oc == 1:
            uo = uos['ua2_object']
        else:
            uo = uos['ua2_object'][objno]
        sec = float(uo['@sec'])
        av = float(uo['@av'])
        pix = int(uo['@pix'])
        bmax = float(uo['@bmax'])
        mag = float(uo['@mag'])
        fcount = int(uo['@sN'])
        return sec, av, pix, bmax, mag, fcount

    def getObjectStart(self, objno):
        uos = self.uaxml['ufoanalyzer_record']['ua2_objects']
        oc = int(self.uaxml['ufoanalyzer_record']['@o'])
        if oc == 1:
            uo = uos['ua2_object']
        else:
            uo = uos['ua2_object'][objno]
        ra1 = uo['@ra1']
        dc1 = uo['@dc1']
        h1 = uo['@h1']
        dist1 = uo['@dist1']
        lng1 = uo['@lng1']
        lat1 = uo['@lat1']
        az1 = uo['@az1']
        ev1 = uo['@ev1']
        fs = uo['@fs']
        return ra1, dc1, h1, dist1, lng1, lat1, az1, ev1, fs

    def getObjectEnd(self, objno):
        uos = self.uaxml['ufoanalyzer_record']['ua2_objects']
        oc = int(self.uaxml['ufoanalyzer_record']['@o'])
        if oc == 1:
            uo = uos['ua2_object']
        else:
            uo = uos['ua2_object'][objno]
        ra2 = uo['@ra2']
        dc2 = uo['@dc2']
        h2 = uo['@h2']
        dist2 = uo['@dist2']
        lng2 = uo['@lng2']
        lat2 = uo['@lat2']
        fe = uo['@fe']
        return ra2, dc2, h2, dist2, lng2, lat2, fe

    def getObjectFrameDetails(self, objno, fno):
        uos = self.uaxml['ufoanalyzer_record']['ua2_objects']
        oc = int(self.uaxml['ufoanalyzer_record']['@o'])
        if oc == 1:
            uo = uos['ua2_object']
        else:
            uo = uos['ua2_object'][objno]
        uop = uo['ua2_objpath']['ua2_fdata2'][fno]
        fno = int(uop['@fno'])
        b = int(uop['@b'])
        lsum = float(uop['@Lsum'])
        mag = float(uop['@mag'])
        az = float(uop['@az'])
        ev = float(uop['@ev'])
        ra = float(uop['@ra'])
        dec = float(uop['@dec'])
        return fno, ra, dec, mag, az, ev, lsum, b

    def getPathVector(self, objno):
        fps, _, _, isintl = UAXml.getCameraDetails(self)
        dtim = UAXml.getDateTime(self)

        uos = self.uaxml['ufoanalyzer_record']['ua2_objects']
        oc = UAXml.getObjectCount(self)
        if oc == 1:
            uo = uos['ua2_object']
        else:
            uo = uos['ua2_object'][objno]
        fcount = int(uo['@sN'])
        fs = int(uo['@fs'])
        fno = numpy.zeros(fcount)
        ra = numpy.zeros(fcount)
        dec = numpy.zeros(fcount)
        alt = numpy.zeros(fcount)
        az = numpy.zeros(fcount)
        tt = numpy.zeros(fcount)
        mag = numpy.zeros(fcount)
        b = numpy.zeros(fcount)
        lsum = numpy.zeros(fcount)
        for fc in range(fcount):
            uop = uo['ua2_objpath']['ua2_fdata2'][fc]
            fno[fc] = int(uop['@fno'])
            ra[fc] = float(uop['@ra'])
            dec[fc] = float(uop['@dec'])
            az[fc] = float(uop['@az'])
            alt[fc] = float(uop['@ev'])
            mag[fc] = float(uop['@mag'])
            b[fc] = float(uop['@b'])
            lsum[fc] = float(uop['@Lsum'])

            # UFO timestamps the first half-frame in which a detection happens and
            # the framenumber of this is stored in "fs"
            us = int((fno[fc] - fs) / fps * 1000000)
            tt[fc] = round((dtim + datetime.timedelta(microseconds=us)).timestamp(), 2)
        return fno, tt, ra, dec, mag, fcount, alt, az, b, lsum


    def makePlateParEntry(self, statid):
        """ Make a fake platepar file from the data, so it can be processed with RMS/WMPL """
        imgdt = self.getDateTime()
        _, cx, cy, _ = self.getCameraDetails()
        az, ev, _, fovh, yx, _, _, _ = self.getProfDetails()
        _, lid, sid, lat, lng, alt = self.getStationDetails()

        ff_name = 'FF_' + statid +'_'
        ff_name = ff_name + imgdt.strftime('%Y%m%d_%H%M%S_') 
        ms = '{:03d}'.format(int(imgdt.microsecond / 1000))
        ff_name = ff_name + ms + '_0000000.fits'

        fovv = fovh*(cx*yx/cy)

        ref_jd = datetime2JD(imgdt)
        ra_d, dec_d = altAz2RADec(numpy.radians(az), numpy.radians(ev), ref_jd, 
            numpy.radians(lat), numpy.radians(lng))

        pp = '"' + ff_name + '": {\n'
        pp = pp + '    "lat": {:f},\n'.format(lat)
        pp = pp + '    "lon": {:f},\n'.format(lng)
        pp = pp + '    "elev": {:f},\n'.format(alt)
        pp = pp + '    "X_res": {:d},\n'.format(int(cx))
        pp = pp + '    "Y_res": {:d},\n'.format(int(cy))
        pp = pp + '    "JD": {:.8f},\n'.format(ref_jd)
        pp = pp + '    "RA_d": {:.8f},\n'.format(numpy.degrees(ra_d))
        pp = pp + '    "dec_d": {:.8f},\n'.format(numpy.degrees(dec_d))
        pp = pp + '    "fov_h": {:f},\n'.format(fovh)
        pp = pp + '    "fov_v": {:f},\n'.format(fovv)
        pp = pp + '    "station_code": "{:s}",\n'.format(statid)
        pp = pp + '    "auto_recalibrated": true,\n'
        pp = pp + '    "az_centre": 0,\n'
        pp = pp + '    "alt_centre": 0,\n'
        pp = pp + '    "rotation_from_horiz": 0,\n'
        pp = pp + '    "pos_angle_ref": 0,\n'
        pp = pp + '    "Ho": 0,\n'
        pp = pp + '    "UT_corr": 0,\n'
        pp = pp + '    "F_scale": 0,\n'
        pp = pp + '    "version": 2\n'
        pp = pp + '}'

        return pp



# Copyright (C) 2018-2023 Mark McIntyre

""" This folder contains some data files used by the library:  
    IMO_Working_Meteor_Shower_List.xml, the IMO's official working list  
    streamfulldata.npy  - a numpy dataset containing the Jenniskens full meteor list, including unconfirmed and disputed showers  

These files are updated whenever the library version is bumped, but if you want to override the files, define an 
environment variable DATADIR, and place your own copies of the files at $DATADIR/share. 
"""


# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa

"""
The functions in this module require RMS and/or WMPL to be in the PythonPath.  

Require RMS, and must be run from the source/RMS folder:  
    multiDayRadiant  
    multiTrackStack  
    analyseUFOwithRMS  

Require WMPL:  
    multiEventGroundMap  
    plotCAMSOrbits  
    plotRMSOrbits  
    getTrajPickle  

"""

from .multiDayRadiant import multiDayRadiant
from .multiTrackStack import multiTrackStack
from .analyseUFOwithRMS import analyseUFOwithRMS
from .multiEventGroundMap import multiEventGroundMap
from .plotCAMSOrbits import plotCAMSOrbits
from .plotRMSOrbits import plotRMSOrbits
from .trajPickle import getTrajPickle, trajectoryKML



# Copyright (C) 2018-2023 Mark McIntyre

#
# python code to plot orbits given a CAMS style Orbit Info file
#

import sys
import os
import datetime
import matplotlib.pyplot as plt
try:
    from wmpl.Utils.PlotOrbits import plotOrbits
except:
    print('WMPL not available')


def plotCAMSOrbits(orbitFile, outdir=None, hideplot=True):
    """
    plots a set of orbits from CAMS data, with one line per set of osculations  

    Arguments:  
        orbitFile:  [string] full path to CAMS orbit details file  
        outdir:     [string] the location to write to. defaults to folder of CAMS file  
        hideplot:   [bool] don't display the plot. Default true  

    Returns:  
        none  
    """
    if outdir is None:
        outdir, _ = os.path.split(orbitFile)
    
    evttime = None

    # read in the data file as a list of lines
    lis = open(orbitFile, 'r').readlines()
    
    # create an empty array to contain the orbital elements
    orb_elements=[]

    # ignore the first two lines, split the remaining lines up and find the 
    # elements we want
    for i in range(3,len(lis)):
        li = lis[i].split()
        if evttime is None:
            evttime = li[1]+'T'+li[2][:8]
        a = float(li[25])
        e = float(li[26])
        if a < 0:
            q = float(li[22])
            if e==1.0:
                e=1.0001
            a = q/(1-e)
        elemline = [a,e,float(li[28]),float(li[30]),float(li[32])]
        # print(elemline)
        # add the elements to the array
        orb_elements.append(elemline)

    # plot and save the orbits
    
    plotOrbits(orb_elements, datetime.datetime.strptime(evttime, '%Y-%m-%dT%H:%M:%S'), color_scheme='light')
    evttime = evttime.replace(':','-')
    plt.savefig(os.path.join(outdir, f'CAMS_{evttime}.png'))
    if not hideplot:
        plt.show()
    plt.close()

#    with open(os.path.join(outdir, 'sampleRMSdata.txt'), 'w') as outf:
#        for li in orb_elements:
#            outf.write(f'{li[0]},{li[1]},{li[2]},{li[3]},{li[4]},{evttime}\n')

    return 


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print('usage: python plotCAMSOrbits source_file target_folder')

    plotCAMSOrbits(sys.argv[1], sys.argv[2])



# Copyright (C) 2018-2023 Mark McIntyre
#
# python code to read in an FTPdetect file and plot all the detections
#

import os
import configparser as cr 
from matplotlib import pyplot as plt


def _readFTPfile(filename, h):
    """ Internal function to load ftpfile into arrays """
    events = []
    with open(filename) as f:
        # Skip the header
        for i in range(11):
            next(f)

        for line in f:
            line = line.replace('\n', '').replace('\r', '')
            # The separator marks the beginning of a new meteor
            if "-------------------------------------------------------" in line:
                # Skip the event header
                for i in range(2):
                    next(f)
            # read line containing detail of this event
            line = f.readline()
            rows = int(line.split()[2])
            xvals=[]
            yvals=[]
            # now read in the data of interest
            for r in range(rows):
                line = f.readline()
                _, x, y, _, _, _, _, _ = list(map(float, line.split()[:8]))
                xvals.append(x)
                yvals.append(h-y)
            dta = {'x': xvals, 'y': yvals}
            events.append(dta)

    return events


def drawFTPFile(ftpfile, cfgfile=None):
    """
    Creates a simple representation of an FTPdetect file, showing the locations of each meteor trail.  
    Its like a stack of the night's detections but in a much simpler format. Useful for diagnosing false detections.   

    Arguments:  
        ftpfile:   [str] full path to the FTPdetect file.  
        cfgfile:   [str] full path to RMS config file to read image dimensions. Default is 1280x720.  
    """
    config = cr.ConfigParser()
    if cfgfile is None: 
        width=1280
        height=720
    else:
        config.read(cfgfile)
        width = int(config['Capture']['width'])
        height = int(config['Capture']['height'])

    print('plotting field of view {}x{}'.format(width, height))
    events = _readFTPfile(ftpfile, height)
    for ev in events: 
        plt.plot(ev['x'],ev['y'])
    pth, ftpf = os.path.split(ftpfile)
    spls = ftpf.split('_')
    outfname = f'{spls[1]}_{spls[2]}_{spls[3]}_{spls[3]}_ftpmap.png'
    plt.savefig(os.path.join(pth, outfname))
    plt.close()
    return 



# Copyright (C) 2018-2023 Mark McIntyre
#
# simple script to get the active shower list from the IMO working list

from ukmon_meteortools.fileformats import imoWorkingShowerList as iwsl
import datetime
import argparse


def getActiveShowers(targdate, retlist=False, inclMinor=False):
    """
    Return a list of showers active at the specified date  

    Arguments:  
        targdate:   [str] Date in YYYYMMDD format  

    Keyword Arguments:  
        retlist:    [bool] return a list, or print to console. Default False=print  
        inclMinor:  [bool] include minor showers or only return major showers  

    Returns:  
        If retlist is true, returns a python list of shower short-codes eg ['PER','LYR']  

    """
    sl = iwsl.IMOshowerList()
    testdate = datetime.datetime.strptime(targdate, '%Y%m%d')
    listofshowers=sl.getActiveShowers(testdate, True, inclMinor=inclMinor)
    if retlist is False:
        for shwr in listofshowers:
            print(shwr)
    else:
        return listofshowers


def getActiveShowersStr(targdatestr):
    """
    Prints a comma-separated list of showers active at the specified date  

    Arguments:  
        targdate:   [str] Date in YYYYMMDD format  

    Returns:  
        nothing  

    """
    shwrs = getActiveShowers(targdatestr, retlist=True)
    shwrs.append('spo')
    for s in shwrs:
        print(s)


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser(description='get list of active showers',
        formatter_class=argparse.RawTextHelpFormatter)
    arg_parser.add_argument('-d', '--targdate', metavar='TARGDATE', type=str,
        help='Date to run for (default is today)')
    arg_parser.add_argument('-m', '--includeminor', action="store_true",
        help='include minor showers')

    cml_args = arg_parser.parse_args()
    if cml_args.targdate is None:
        targdate = datetime.datetime.now().strftime('%Y%m%d')
    getActiveShowers(targdate, inclMinor=cml_args.includeminor)



# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa
"""
Utility functions used across the UKMON toolset

List of Functions:  
jd2Date, date2JD, datetime2JD, jd2DynamicalTimeJD, jd2LST, sollon2jd  
greatCircleDistance, angleBetweenSphericalCoords, calcApparentSiderealEarthRotation  
calcNutationComponents, equatorialCoordPrecession,  raDec2AltAz, altAz2RADec  
altAz2RADec_vect, raDec2AltAz_vect, equatorialCoordPrecession_vect  
annotateImage, annotateImageArbitrary  
getActiveShowers, getActiveShowersStr, getShowerDets, getShowerPeak  
sendAnEmail  
shortestDistance2Lines  
"""

from .Math import jd2Date, date2JD,datetime2JD, jd2DynamicalTimeJD, JULIAN_EPOCH, J2000_JD, jd2LST
from .Math import greatCircleDistance, angleBetweenSphericalCoords, calcApparentSiderealEarthRotation
from .Math import calcNutationComponents, equatorialCoordPrecession,  raDec2AltAz, altAz2RADec
from .Math import altAz2RADec_vect, raDec2AltAz_vect, equatorialCoordPrecession_vect
from .annotateImage import annotateImage, annotateImageArbitrary
from .convertSolLon import sollon2jd
from .getActiveShowers import getActiveShowers, getActiveShowersStr
from .getShowerDates import getShowerDets, getShowerPeak
from .sendAnEmail import sendAnEmail
from .drawFTPfile import drawFTPFile



# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa
"""
Functions to load and manage various meteor file formats

"""

from .ftpDetectInfo import filterFTPforSpecificTime, writeNewFTPFile, loadFTPDetectInfo, MeteorObservation
from .imoWorkingShowerList import IMOshowerList, majorlist, minorlist
from .platepar import loadPlatepars, platepar
from .UFOAnalyzerXML import UAXml
from .UFOCapXML import UCXml
from .kmlHandlers import trackCsvtoKML, getTrackDetails, munchKML
from .ECSVhandler import getECSVs


# Copyright (C) 2018-2023 Mark McIntyre
# 
#
import os
import glob


from ukmon_meteortools.fileformats import munchKML
from shapely.geometry import Point


def pointInsideFov(latDegs,lngDegs, kmlFilename):
    """
    Test if a point is inside the field of view of a KML file  

    Arguments:  
        latDegs:        [degrees] latitude of the point  
        lngDegs:        [degrees] latitude of the point  
        kmlFileName:    [str] full path to the KML file to test  

    Returns:  
        True or False  
    """

    c1, p1 = munchKML(kmlFilename, True)
    pt = Point(latDegs, lngDegs)
    return p1.contains(pt)


def checkKMLOverlap(kmfile1, kmfile2):
    """
    Test if two KML files overlap  

    Arguments:  
        kmlfile1:       [str] full path to first KML file  
        kmlfile2:       [str] full path to second KML file  

    Returns:  
        True or False
    """
    _,p1 = munchKML(kmfile1, True)
    _,p2 = munchKML(kmfile2, True)
    return p1.intersects(p2)


def getOverlapWith(srcfolder, kmlpattern='*-25km.kml', refcam='UK0006'):
    """
    Check for overlap between the named camera, and every KML file in a folder, at the pattern altitude  

    Arguments:  
        srcfolder:  [str] path to folder containing KML files to test.   
        kmlpattern: [str] kml pattern to match. Default "*-25km.kml"  
        refcam:     [str] Cameraid to check against. Its KML must be in srcfolder  
     
    Returns:  
        list of camera IDs that overlap with the target  
    """
    kmllist = glob.glob1(srcfolder, kmlpattern)
    currmatches=[]
    refkml = f'{refcam}{kmlpattern[1:]}'
    currmatches.append(refcam[:6])
    print('checking ', refkml)
    for testkml in kmllist:
        if testkml != refkml:
            testcam,_ = os.path.splitext(testkml)
            #print(testkml)
            #print('comparing to ', testcam)
            if checkKMLOverlap(os.path.join(srcfolder, refkml), os.path.join(srcfolder, testkml)) is True:
                currmatches.append(testcam[:6])
    return currmatches



def getOverlappingCameras(srcfolder, kmlpattern='*-25km.kml'):
    """
    Check for all overlaps in the folder at the pattern altitude  

    Arguments:  
        srcfolder:  [str] path to folder containing KML files to test.   
        kmlpattern: [str] kml pattern to match. Default "*-25km.kml"  

    Returns:  
        list of lists of groups of overlaping IDs 
     
    """
    kmllist = glob.glob1(srcfolder, kmlpattern)
    kmllist2 = kmllist
    matches = []

    for refkml in kmllist:
        currmatches=[]
        refcam,_ = os.path.splitext(refkml)
        currmatches.append(refcam[:6])
        #print('checking ', refcam)
        for testkml in kmllist2:
            if testkml != refkml:
                testcam,_ = os.path.splitext(testkml)
                #print('comparing to ', testcam)
                if checkKMLOverlap(os.path.join(srcfolder, refkml), os.path.join(srcfolder, testkml)) is True:
                    currmatches.append(testcam[:6])
        matches.append(currmatches)
    return matches



import xmltodict
from shapely.geometry import Polygon
import csv
import simplekml
import numpy as np
import pandas as pd
import os


def munchKML(kmlFilename, return_poly=False):
    """ Load a KML file and return either a list of lats and longs, or a Shapely polygon  
    
    Arguments:  
        kmlFilename:    [string] full path to the KML file to consume   
        return_poly:    [bool] return a Shapely polygon? Default False  

    
    Returns:  
        if return_poly is false, returns a tuple of (cameraname, lats, longs) where lats and longs are lists of the 
        latitudes and longitudes in the KML file.  

        If return_poly is true, returns a tuple of (cameranamem, shapely Polygon)  
    """

    with open(kmlFilename) as fd:
        x = xmltodict.parse(fd.read())
        cname = x['kml']['Folder']['name']
        coords = x['kml']['Folder']['Placemark']['MultiGeometry']['Polygon']['outerBoundaryIs']['LinearRing']['coordinates']
        coords = coords.split('\n')
        if return_poly is False:
            lats = []
            lngs = []
            for lin in coords:
                s = lin.split(',')
                lngs.append(float(s[0]))
                lats.append(float(s[1]))
            return cname, lats, lngs
        else:
            ptsarr=[]
            for lin in coords:
                s = lin.split(',')
                ptsarr.append((float(s[0]), float(s[1])))
            polyg = Polygon(ptsarr)
            return cname, polyg 


def trackCsvtoKML(trackcsvfile, trackdata=None, saveOutput=True, outdir=None):
    """ 
    Either reads a CSV file containing lat, long, height of an event and 
    creates a 3d KML file from it or, if trackdata is populated, converts a Pandas dataframe containing 
    the same data. Output is written to disk unless saveOutput is false.  
    
    Arguments:  
        trackcsvfile:   [string] full path to the file to read from  
        trackdata:      [array] pandas dataframe containing the data. Default None  
        saveOutput:     [bool] write the KML file to disk. Default true  
        outdir:         [string] where to save the file. Default same folder as the source file

    Returns:  
        the KML file as a tuple  
        """
    kml=simplekml.Kml()
    kml.document.name = trackcsvfile
    if trackdata is None:
        inputfile = csv.reader(open(trackcsvfile))
        for row in inputfile:
            #columns are lat, long, height, times
            kml.newpoint(name='', coords=[(row[1], row[0], row[2])])
    else:
        for i,r in trackdata.iterrows():
            kml.newpoint(name='', coords=[(r[1], r[0], r[2])], extrude=1, altitudemode='absolute')
    if 'csv' in trackcsvfile:
        outname = trackcsvfile.replace('.csv','.kml')
    else:
        outname = f'{trackcsvfile}.kml'
    if saveOutput:
        if outdir is None:
            outdir, _ = os.path.split(trackcsvfile)
        os.makedirs(outdir, exist_ok=True)
        outname = os.path.join(outdir, outname)
        kml.save(outname)
    return kml


def getTrackDetails(traj):
    """ Get track details from a WMPL trajectory object  
    
    Arguments:  
        traj:       a WMPL trajectory object containing observations  

    Returns:  
        a Pandas dataframe containing the lat, long, alt and time of each point on the trajectory, sorted by time. 
    """
    lats = []
    lons = []
    alts = [] 
    lens = []
    # Go through observation from all stations
    for obs in traj.observations:
        # Go through all observed points
        for i in range(obs.kmeas):
            lats.append(np.degrees(obs.model_lat[i]))
            lons.append(np.degrees(obs.model_lon[i]))
            alts.append(obs.model_ht[i])
            lens.append(obs.time_data[i])
    df = pd.DataFrame({"lats": lats, "lons": lons, "alts": alts, "times": lens})
    df = df.sort_values(by=['times', 'lats'])
    return df



# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa

from .apiExampleCode import matchApiCall, detailApiCall1, detailApiCall2
from .findNearDuplicates import findNearDuplicates
from .getLiveImages import getLiveJpgs, getFBfiles, createTxtFile
from .getOverlappingFovs import checkKMLOverlap, pointInsideFov, getOverlapWith, getOverlappingCameras
from .plotTrack import trackToDistvsHeight, trackToTimevsVelocity, trackToTimevsHeight



# Copyright (C) 2018-2023 Mark McIntyre

import os
import glob
import datetime
try:
    from Utils.ShowerAssociation import showerAssociation
    import RMS.ConfigReader as cr
except Exception:
    print('RMS not available')
import argparse
import shutil


def multiDayRadiant(camlist, start, end, outdir=None, shwr=None, datadir=None):
    """
    Create a multi-day and/or multi-camera radiant map for cameras at the same geographic location. 

    Arguments:  
        camlist:    [list] list of camera IDs eg ['UK0006','UK000F']  
        start:      [string] start date in YYYYMMDD format  
        end:        [string] end date in YYYYMMDD format  
        outdir:     [string] where to save the file to. Default is current directory  
        shwr:       [string] Filter by shower eg PER. Default all showers.   
        datadir:    [string] Where to look for the data. Default f:/videos/meteorcam.  

    Notes:  
        The function expects data to be stored in RMS folder structures as follows  
            {datadir}/{cameraid}/ConfirmedFiles/{cameraid_date_time_*}  

        Each folder must contain an FTPdetectinfo file and a platepars_all file. 
        The first-named folder must contain a valid RMS config file.  

        It is assumed that all cameras are at the same location. The output for cameras 
        at different locations has not been tested.  

    Output:
        creates two files in outdir, a PNG containing an all-sky image of the radiants and 
        meteor tracks, and a text file containing the same information

    """
    if datadir is None:
        locfld = os.getenv('LOCALFOLDER', default='f:/videos/meteorcam/fireballs')
        datadir, _ = os.path.split(locfld)
    if outdir is None:
        outdir = '.'
    print(f'stacking {camlist} for dates {start},{end}, reading from {datadir}, saving to {outdir}')

    sdt = datetime.datetime.strptime(start, '%Y%m%d')
    edt = datetime.datetime.strptime(end, '%Y%m%d')
    d = sdt
    dir_paths = []
    while d <= edt:
        dtstr = d.strftime('%Y%m%d')
        for cam in camlist:
            cam = cam.upper()
            camdir = os.path.join(datadir, cam, 'ConfirmedFiles')
            dirs = glob.glob1(camdir, f'{cam}_{dtstr}*')
            if len(dirs) > 0: 
                srcdir = os.path.join(camdir, dirs[0])
                ftpdet = os.path.join(srcdir, f'FTPdetectinfo_{dirs[0]}.txt')
                if os.path.isfile(ftpdet):
                    print(f'using {ftpdet}')
                    dir_paths.append(ftpdet)
        d = d + datetime.timedelta(days=1)
    cfgdir, _ = os.path.split(dir_paths[0])
    config = cr.loadConfigFromDirectory('.config', cfgdir)

    # file will be created here, need to preserve any existing file then restore afterwards
    targdir, _ = os.path.split(dir_paths[-1])
    prevfiles = glob.glob1(targdir, '*radiants.png')
    if len(prevfiles) > 0:
        if os.path.isfile(os.path.join(targdir, prevfiles[0]+'.tmp')):
            os.remove(os.path.join(targdir, prevfiles[0]+'.tmp'))
        os.rename(os.path.join(targdir, prevfiles[0]), os.path.join(targdir, prevfiles[0]+'.tmp'))
    prevfilestxt = glob.glob1(targdir, '*radiants.txt')
    if len(prevfilestxt) > 0:
        if os.path.isfile(os.path.join(targdir, prevfilestxt[0]+'.tmp')):
            os.remove(os.path.join(targdir, prevfilestxt[0]+'.tmp'))
        os.rename(os.path.join(targdir, prevfilestxt[0]), os.path.join(targdir, prevfilestxt[0]+'.tmp'))

    # Perform shower association
    associations, shower_counts = showerAssociation(config, dir_paths, 
        shower_code=shwr, show_plot=False, save_plot=True, plot_activity=True,
        flux_showers=False, color_map='gist_ncar')

    prevfiles = glob.glob1(targdir, '*radiants.png')
    prevfilestxt = glob.glob1(targdir, '*radiants.txt')
    newn = prevfiles[0]
    newntxt = prevfilestxt[0]
    if shwr is not None:
        newn = shwr + newn[6:]
        newntxt = shwr + newntxt[6:]
    else:
        newn = "ALL" + newn[6:]
        newntxt = "ALL" + newntxt[6:]
    shutil.copyfile(os.path.join(targdir, prevfiles[0]), os.path.join(outdir, newn))
    shutil.copyfile(os.path.join(targdir, prevfilestxt[0]), os.path.join(outdir, newntxt))
    os.remove(os.path.join(targdir, prevfiles[0]))
    os.remove(os.path.join(targdir, prevfilestxt[0]))

    prevfiles = glob.glob1(targdir, '*radiants.png')
    if len(prevfiles) > 0:
        os.rename(os.path.join(targdir, prevfiles[0])+'.tmp', os.path.join(targdir, prevfiles[0]))

    prevfilestxt = glob.glob1(targdir, '*radiants.txt')
    if len(prevfilestxt) > 0:
        os.rename(os.path.join(targdir, prevfilestxt[0])+'.tmp', os.path.join(targdir, prevfilestxt[0]))
    return


if __name__ == '__main__':

    arg_parser = argparse.ArgumentParser(description="Perform multiday/camera radiant mapping.")

    arg_parser.add_argument('cams', metavar='CAMS', nargs='+', type=str,
        help='List of cameras, comma-separated.')

    arg_parser.add_argument('dates', metavar='DATES', nargs='+', type=str,
        help='Start and end date, comma-separated in yyyymmdd format.')

    arg_parser.add_argument('-s', '--shower', metavar='SHOWER', type=str,
        help="Associate just this single shower given its code (e.g. PER, ORI, ETA).")

    arg_parser.add_argument('-o', '--outdir', metavar='OUTDIR', type=str,
        help="Where to save the output file.")

    # Parse the command line arguments
    cml_args = arg_parser.parse_args()
    cams = cml_args.cams[0]
    if ',' in cams:
        cams = cams.split(',')
    else:
        cams = [cams]

    dates = cml_args.dates[0]
    if ',' in dates:
        start, end = dates.split(',')
    else:
        start = dates
        end = start
    
    shwr = None
    if cml_args.shower:
        shwr = cml_args.shower
    
    outdir = '.'
    if cml_args.outdir:
        outdir = cml_args.outdir

    multiDayRadiant(cams, start, end, outdir, shwr)



# Copyright (C) 2018-2023 Mark McIntyre

import datetime
import pandas as pd
import argparse
import os
import matplotlib.pyplot as plt
try:
    from wmpl.Utils.PlotMap_OSM import OSMMap
except:
    print('WMPL not available')

RAD2DEG=57.2958


def multiEventGroundMap(startdt, enddt, statid=None, shwr=None, outdir=None):
    """
    Plots a ground track diagram of all events between two dates, with filters by station and shower  

    Arguments:  
        start:      [string] start date in YYYYMMDD format  
        end:        [string] end date in YYYYMMDD format  
        statid:     [string] station to filter for. Default all stations.  
        shwr:       [string] Filter by shower eg PER. Default All showers.  
        outdir:     [string] where to save the file to. if this parameter is omitted, the image will be displayed not saved  

    Output:  
        A jpg map of the detections.   

    Note:  
        This function reads directly from the UKMON public dataset.  

    """
    yr = startdt[:4]

    cols=['_lat1', '_lng1','_lat2','_lng2','_stream','dtstamp', 'stations']
    matchfile = f'https://archive.ukmeteornetwork.co.uk/browse/parquet/matches-full-{yr}.parquet.snap'

    dta = pd.read_parquet(matchfile, columns=cols)

    # filter the data down to just the cols we want
    sd = datetime.datetime.strptime(startdt, '%Y%m%d')
    sd = sd+datetime.timedelta(hours=12)
    if startdt == enddt:
        ed = sd + datetime.timedelta(days=1)
    else:
        ed = datetime.datetime.strptime(enddt, '%Y%m%d')
        ed = ed+datetime.timedelta(hours=12)
    dta = dta[dta.dtstamp >= sd.timestamp()]
    dta = dta[dta.dtstamp <= ed.timestamp()]
    
    if shwr is not None:
        dta = dta[dta._stream==shwr.upper()]
    if statid is not None:
        dta = dta[dta.stations.str.contains(statid.upper())]

    if len(dta) > 1:
        lat_list = [min(min(dta._lat1), min(dta._lat2))/RAD2DEG, max(max(dta._lat1), max(dta._lat2))/RAD2DEG]
        lon_list = [min(min(dta._lng1), min(dta._lng2))/RAD2DEG, max(max(dta._lng1), max(dta._lng2))/RAD2DEG]
        # Init the map
        #print(lat_list, lon_list)
        m = OSMMap(lat_list, lon_list, border_size=50, color_scheme='dark')
        lat1s=list(dta._lat1)
        lat2s=list(dta._lat2)
        lng1s=list(dta._lng1)
        lng2s=list(dta._lng2)
        for l1, l2, g1, g2 in zip(lat1s, lat2s, lng1s, lng2s):
            lats = [l1/RAD2DEG, l2/RAD2DEG]
            lons = [g1/RAD2DEG, g2/RAD2DEG]
            m.plot(lats, lons, c='r')
            m.scatter(l2/RAD2DEG, g2/RAD2DEG, c='k', marker='+', s=50, alpha=0.75)
        if outdir is not None:
            plt.savefig(os.path.join(outdir, f'{startdt}-{enddt}-{shwr}-{statid}.jpg'))
        else:
            plt.show()
        plt.close()
    return 


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser(description="""Plot a ground map of many detections.""",
        formatter_class=argparse.RawTextHelpFormatter)

    arg_parser.add_argument('start_date', type=str, help='start date in yyyymmdd format')
    arg_parser.add_argument('end_date', type=str, help='end date in yyyymmdd format')
    arg_parser.add_argument('-s', '--shower', metavar='SHOWER', type=str,
        help="Map just this single shower given its code (e.g. PER, ORI, ETA).")

    arg_parser.add_argument('-i', '--stationid', metavar='STATID', help='Station id eg UK0006')
    arg_parser.add_argument('-o', '--outdir', metavar='OUTDIR', help='Location to save jpg into')

    cml_args = arg_parser.parse_args()

    multiEventGroundMap(cml_args.start_date, cml_args.end_date, 
        cml_args.stationid, cml_args.shower, cml_args.outdir)



# Copyright (C) 2018-2023 Mark McIntyre

#
# python code to plot orbits given a CAMS style Orbit Info file
#

import sys
import os
import pandas as pd
import datetime
import matplotlib.pyplot as plt
try:
    from wmpl.Utils.PlotOrbits import plotOrbits
except:
    print('WMPL not available')


def plotRMSOrbits(orbitFile, outdir=None, hideplot=True):
    """
    plots a set of orbits from RMS data, with one line per set of osculations  

    Arguments:  
        orbitFile:  [string] full path to RMS orbit details file   
        outdir:     [string] the location to write to. defaults to folder of source file  
        hideplot:   [bool] don't display the plot. Default true  

    Returns:  
        none  

    Notes:   
        Orbitfile should be a csv file with the following labelled columns:  
            _a, _e, _incl, _peri, _node, _datetime  

        _datetime should be in the format YYYY-MM-DDTHH-MM-SS

    """
    if outdir is None:
        outdir, _ = os.path.split(orbitFile)

    evttime = None

    # read in the data file as a list of lines
    data = pd.read_csv(orbitFile)
    avals = list(data['_a'])
    evals = list(data['_e'])
    ivals = list(data['_incl'])
    wvals = list(data['_peri'])
    nvals = list(data['_node'])
    dvals = list(data['_datetime'])
    # create an empty array to contain the orbital elements
    orb_elements=[]

    # ignore the first two lines, split the remaining lines up and find the 
    # elements we want
    for i in range(len(avals)):
        if avals[i] > 0 and avals[i]<=10:
            elemline = [avals[i],evals[i],ivals[i],wvals[i],nvals[i]]
            # add the elements to the array
            orb_elements.append(elemline)
        if evttime is None:
            evttime = dvals[i]

    # plot and save the orbits
    plotOrbits(orb_elements, datetime.datetime.strptime(evttime, '%Y-%m-%dT%H-%M-%S'), color_scheme='light')
    plt.savefig(os.path.join(outdir, f'RMS_{evttime}.png'))
    if not hideplot:
        plt.show()
    plt.close()

    return 


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print('usage: python plotRMSOrbits source_file target_folder')

    plotRMSOrbits(sys.argv[1], sys.argv[2])



# Copyright (C) 2018-2023 Mark McIntyre

from PIL import Image, ImageFont, ImageDraw 
import datetime


def annotateImage(img_path, statid, metcount, rundate=None):
    """
    Annotate an image with the station ID and date in the bottom left and meteor count in the  
    bottom right  

    Arguments:  
        img_path:   [str] full path and filename of the image to be annotated  
        statid:     [str] station ID string to use  
        metcount:   [int] number of meteors in the image  

    
    Keyword Args:  
        rundate:    [str] rundate in 'YYYYMM' or 'YYYYMMDD' format. Default is today.   

    """
    if rundate is not None:
        if len(rundate) > 6:
            now = datetime.datetime.strptime(rundate, '%Y%m%d')
            title = '{} {}'.format(statid, now.strftime('%Y-%m-%d'))
        else:
            now = datetime.datetime.strptime(rundate, '%Y%m')
            title = '{} {}'.format(statid, now.strftime('%Y-%m'))
    else:
        now = datetime.datetime.now()
        title = '{} {}'.format(statid, now.strftime('%Y-%m-%d'))

    my_image = Image.open(img_path)
    width, height = my_image.size
    image_editable = ImageDraw.Draw(my_image)
    fntheight=30
    try:
        fnt = ImageFont.truetype("arial.ttf", fntheight)
    except:
        fnt = ImageFont.truetype("DejaVuSans.ttf", fntheight)
    #fnt = ImageFont.load_default()
    image_editable.text((15,height-fntheight-15), title, font=fnt, fill=(255))
    metmsg = 'meteors: {:04d}'.format(metcount)
    image_editable.text((width-7*fntheight-15,height-fntheight-15), metmsg, font=fnt, fill=(255))
    my_image.save(img_path)


def annotateImageArbitrary(img_path, message, color='#000'):
    """
    Annotate an image with an arbitrary message in the selected colour at the bottom left  

    Arguments:  
        img_path:   [str] full path and filename of the image to be annotated  
        message:    [str] message to put on the image  
        color:      [str] hex colour string, default '#000' which is black  

    """
    my_image = Image.open(img_path)
    width, height = my_image.size
    image_editable = ImageDraw.Draw(my_image)
    fntheight=30
    try:
        fnt = ImageFont.truetype("arial.ttf", fntheight)
    except:
        fnt = ImageFont.truetype("DejaVuSans.ttf", fntheight)
    #fnt = ImageFont.load_default()
    image_editable.text((15,height-fntheight-15), message, font=fnt, fill=color)
    my_image.save(img_path)



#
# Example showing how to use the matchdata API from Python
#
# Copyright (C) 2018-2023 Mark McIntyre

"""
Examples showing how to use the APIs. 
"""
import json
import requests 
import pandas as pd


def matchApiCall(reqtyp, reqval):
    """get names of all matches for a given date"""
    apiurl = 'https://api.ukmeteornetwork.co.uk/matches'
    apicall = f'{apiurl}?reqtyp={reqtyp}&reqval={reqval}'
    matchlist = pd.read_json(apicall, lines=True)
    return matchlist


def getMatchPickle(patt):
    """
    get the pickled trajectory for a given matched event   
    Returns a json object containing the pickled data
    
    Example pattern 20230501_002536.754_UK
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    apicall = f'{apiurl}?reqval={patt}'
    res = requests.get(apicall, stream=True)
    if res.status_code == 200:
        data=b''
        for chunk in res.iter_content(chunk_size=4096):
            data = data + chunk
        jsd = json.loads(data)  
    else:
        jsd = None  
    return jsd


def detailApiCall1(reqtyp, reqval):
    """ get details of one event """
    apiurl = 'https://api.ukmeteornetwork.co.uk/matches'
    apicall = f'{apiurl}?reqtyp={reqtyp}&reqval={reqval}'
    evtdetail = pd.read_json(apicall, typ='series')
    return evtdetail


def detailApiCall2(reqtyp, matchlist):
    """
    get details for the first five events in the match list 
    and put them in a pandas dataframe, then sort by brightest
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/matches'
    details = []
    for id in matchlist.head(5).orbname:
        reqval = id
        apicall = f'{apiurl}?reqtyp={reqtyp}&reqval={reqval}'
        details.append(pd.read_json(apicall, typ='series'))
    df = pd.DataFrame(details)
    df = df.sort_values(by=['_mag'])
    return df


def getLiveimageList(dtstr):
    """ 
    Get a list of livestream images matching a pattern YYYYMMDD_HHMMSS.  
    The seconds and minutes parts are optional but huge amounts of data may get returned.  

    Example pattern: '20230421_2122'
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/liveimages/getlive'
    liveimgs = pd.read_json(f'{apiurl}?pattern={dtstr}')
    return liveimgs


def getFireballFiles(patt):
    """ 
    Get a zip file containing the fireball data matching a pattern of the form UKxxxx_YYYYMMDD_HHMMSS
    Nothing will be returned if there is no fireball data available. 

    Example pattern: 'UK0006_20230421_2122'

    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/fireballs/getfb'
    fbfiles = pd.read_json(f'{apiurl}?pattern={patt}')
    return fbfiles


def trajectoryAPI(trajname):
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    fmt = 'json'
    apicall = f'{apiurl}?reqval={trajname}&format={fmt}'
    res = requests.get(apicall, stream=True)
    if res.status_code == 200:
        data=b''
        for chunk in res.iter_content(chunk_size=4096):
            data = data + chunk
        jsd = json.loads(data)  
    else:
        jsd = None  
    return jsd



# Copyright (C) 2018-2023 Mark McIntyre

import pandas as pd


def findNearDuplicates():
    cols=['_mjd','_localtime','_lat1','_lng1','_lat2','_lng2','numstats','stations']
    df = pd.read_parquet('matched/matches-full-2022.parquet.snap', columns=cols)
    df['l1diff']=df._lat1.diff()*60
    df['l2diff']=df._lat2.diff()*60
    df['g1diff']=df._lng1.diff()*60
    df['g2diff']=df._lng2.diff()*60
    df['dtdiff']=df._mjd.diff()*86400
    df['stdiff']=df.numstats.diff()
    df['bl1diff']=df._lat1.diff(periods=-1)*60
    df['bl2diff']=df._lat2.diff(periods=-1)*60
    df['bg1diff']=df._lng1.diff(periods=-1)*60
    df['bg2diff']=df._lng2.diff(periods=-1)*60
    df['bdtdiff']=df._mjd.diff(periods=-1)*86400
    df['bstdiff']=df.numstats.diff(periods=-1)
    df = df.fillna(999)

    df2 = df[(abs(df.dtdiff) <= 1.0) | (abs(df.bdtdiff) <= 1.0)]
    df2 = df2[(abs(df2.l1diff) <= 2) | (abs(df2.bl1diff) <= 2)]
    df2 = df2[(abs(df2.l2diff) <= 2) | (abs(df2.bl2diff) <= 2)]
    df2 = df2[(abs(df2.g1diff) <= 2) | (abs(df2.bg1diff) <= 2)]
    df2 = df2[(abs(df2.g2diff) <= 2) | (abs(df2.bg2diff) <= 2)]

    numsame=0
    for i,rw in df2.iterrows():
        if df.iloc[i-1].stations != rw.stations:
            numsame +=1
            print(rw._localtime)
        print(f' number with the same cameras:  {numsame}')
        print(f' number with different cameras: {len(df2)-numsame}')
